#!/usr/bin/env python
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Paul Nilsson, paul.nilsson@cern.ch, 2018

import logging
logger = logging.getLogger(__name__)


def run():
    """
    Dummy run function.

    :return:
    """

    pass



#!/usr/bin/env python
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Wen Guan, wen.guan@cern,ch, 2018

import traceback
import logging

from pilot.info import infosys
from pilot.common.exception import PilotException, ErrorCodes
from pilot.api.data import StagingClient, StageInClient, StageOutClient

logger = logging.getLogger(__name__)


class StagingESClient(StagingClient):
    """
        Base ES Staging Client
    """

    def __init__(self, infosys_instance=None, acopytools=None, logger=logger, default_copytools='rucio', default_activity='default', trace_report=None):
        """
            If `acopytools` is not specified then it will be automatically resolved via infosys. In this case `infosys` requires initialization.
            :param acopytools: dict of copytool names per activity to be used for transfers. Accepts also list of names or string value without activity passed.
            :param logger: logging.Logger object to use for logging (None means no logging)
            :param default_copytools: copytool name(s) to be used in case of unknown activity passed. Accepts either list of names or single string value.
            "param default_activity: default activity name
        """

        super(StagingESClient, self).__init__(infosys_instance=infosys_instance, acopytools=acopytools, logger=None, default_copytools=default_copytools,
                                              trace_report=trace_report)

        self.copytool_modules['objectstore'] = {'module_name': 'objectstore'}

        if 'es_events' not in self.acopytools:
            self.acopytools['es_events'] = ['objectstore']
        if 'es_events_read' not in self.acopytools:
            self.acopytools['es_events_read'] = ['objectstore']

        self.astorages = {}
        if self.infosys and self.infosys.queuedata and self.infosys.queuedata.astorages:
            self.astorages = (self.infosys.queuedata.astorages or {}).copy()

        # es_events_read should be able to read from all storages
        if 'es_events_read' not in self.astorages:
            self.astorages['es_events_read'] = []

        logger.info('Configured astorages per activity: astorages=%s' % self.astorages)

    def transfer(self, files, activity=['pw'], **kwargs):  # noqa: C901
        """
            Automatically stage passed files using copy tools related to given `activity`
            :param files: list of `FileSpec` objects
            :param activity: list of activity names used to determine appropriate copytool (prioritized list)
            :param kwargs: extra kwargs to be passed to copytool transfer handler
            :raise: PilotException in case of controlled error
            :return: output of copytool trasfers (to be clarified)
        """

        if isinstance(activity, basestring):
            activity = [activity]

        result, errors = None, []
        avail_activity = False
        for act in activity:
            copytools = self.acopytools.get(act)
            storages = self.astorages.get(act)
            if not copytools:
                logger.warn("No available copytools for activity %s" % act)
                continue
            if act in ['pw', 'pls', 'es_events', 'es_failover'] and not storages:
                # for write activity, if corresponding storages are not defined, should use different activity
                logger.warn("Failed to find corresponding astorages for writing activity(%s), will try next activity" % act)
                continue

            storage = storages[0] if storages else None
            avail_activity = True
            for name in copytools:
                try:
                    if name not in self.copytool_modules:
                        raise PilotException('passed unknown copytool with name=%s .. skipped' % name)
                    module = self.copytool_modules[name]['module_name']
                    logger.info('Trying to use copytool=%s for activity=%s' % (name, act))
                    copytool = __import__('pilot.copytool.%s' % module, globals(), locals(), [module], -1)
                except PilotException as e:
                    errors.append(e)
                    logger.debug('Error: %s' % e)
                    continue
                except Exception as e:
                    logger.warning('Failed to import copytool module=%s, error=%s' % (module, e))
                    logger.debug('Error: %s' % e)
                    continue

                try:
                    result = self.transfer_files(copytool, files, act, storage, **kwargs)
                except PilotException, e:
                    errors.append(e)
                    logger.debug('Error: %s' % e)
                except Exception as e:
                    logger.warning('Failed to transfer files using copytool=%s .. skipped; error=%s' % (copytool, e))
                    logger.error(traceback.format_exc())
                    errors.append(e)

                if errors and isinstance(errors[-1], PilotException) and errors[-1].get_error_code() == ErrorCodes.MISSINGOUTPUTFILE:
                    raise errors[-1]

                if result:
                    break
                else:
                    logger.warn("Failed to transfer files using activity(%s) copytool(%s) with error=%s" % (act, name, errors))
            if result:
                break
            else:
                logger.warn("Failed to transfer files using activity(%s) with copytools(%s)" % (act, copytools))

        if not avail_activity:
            raise PilotException('Not available activity with both acopytools and astorages defined')
        if not result:
            raise PilotException('Failed to transfer files with activities %s' % (activity))

        return result


class StageInESClient(StagingESClient, StageInClient):

    def process_storage_id(self, files):
        """
        If storage_id is specified, replace ddmendpoint by parsing storage_id
        """
        for fspec in files:
            if fspec.storage_token:
                storage_id, path_convention = fspec.get_storage_id_and_path_convention()
                if path_convention and path_convention == 1000:
                    fspec.scope = 'transient'
                if storage_id:
                    fspec.ddmendpoint = infosys.get_ddmendpoint(storage_id)
                logger.info("Processed file with storage id: %s" % fspec)

    def transfer_files(self, copytool, files, activity, ddmendpoint, **kwargs):
        """
            Automatically stage in files using the selected copy tool module.

            :param copytool: copytool module
            :param files: list of `FileSpec` objects
            :param activity: activity name.
            :param ddmendpoint: ddmendpoint name.
            :param kwargs: extra kwargs to be passed to copytool transfer handler

            :return: the output of the copytool transfer operation
            :raise: PilotException in case of controlled error
        """
        self.process_storage_id(files)
        return super(StageInESClient, self).transfer_files(copytool, files, activity=activity, ddmendpoint=ddmendpoint, **kwargs)


class StageOutESClient(StagingESClient, StageOutClient):

    def transfer_files(self, copytool, files, activity, ddmendpoint, **kwargs):
        """
            Automatically stage out files using the selected copy tool module.

            :param copytool: copytool module
            :param files: list of `FileSpec` objects
            :param activity: activity name.
            :param ddmendpoint: ddmendpoint name.
            :param kwargs: extra kwargs to be passed to copytool transfer handler

            :return: the output of the copytool transfer operation
            :raise: PilotException in case of controlled error
        """
        logger.info("To transfer files with activity: %s" % (activity))
        for fspec in files:
            if not fspec.ddmendpoint or fspec.ddmendpoint != ddmendpoint:
                logger.info("Based on activity %s, changing ddmendpoint from '%s' for file(%s) to '%s'" % (activity, fspec.ddmendpoint, fspec.lfn, ddmendpoint))
                fspec.ddmendpoint = ddmendpoint

        return super(StageOutESClient, self).transfer_files(copytool, files, activity, **kwargs)



#!/usr/bin/env python
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Paul Nilsson, paul.nilsson@cern.ch, 2017-2019

import os
import pipes
import re
# for user container test: import urllib

from pilot.user.atlas.setup import get_asetup
from pilot.user.atlas.setup import get_file_system_root_path
from pilot.info import infosys
from pilot.util.auxiliary import get_logger
from pilot.util.config import config
from pilot.util.filehandling import write_file
# import pilot.info.infoservice as infosys

import logging
logger = logging.getLogger(__name__)


def do_use_container(**kwargs):
    """
    Decide whether to use a container or not.

    :param kwargs: dictionary of key-word arguments.
    :return: True if function has decided that a container should be used, False otherwise (boolean).
    """

    use_container = False

    # return use_container

    # to force no container use: return False

    job = kwargs.get('job')
    if job:
        # for user jobs, TRF option --containerImage must have been used, ie imagename must be set
        if job.is_analysis() and job.imagename:
            use_container = False
        else:
            queuedata = job.infosys.queuedata
            container_name = queuedata.container_type.get("pilot")
            if container_name == 'singularity':
                use_container = True

    return use_container


def wrapper(executable, **kwargs):
    """
    Wrapper function for any container specific usage.
    This function will be called by pilot.util.container.execute() and prepends the executable with a container command.

    :param executable: command to be executed (string).
    :param kwargs: dictionary of key-word arguments.
    :return: executable wrapped with container command (string).
    """

    workdir = kwargs.get('workdir', '.')
    pilot_home = os.environ.get('PILOT_HOME', '')
    job = kwargs.get('job')

    logger.info('container wrapper called')

    if workdir == '.' and pilot_home != '':
        workdir = pilot_home

    # if job.imagename (from --containerimage <image>) is set, then always use raw singularity
    if config.Container.setup_type == "ALRB" and not job.imagename:
        fctn = alrb_wrapper
    else:
        fctn = singularity_wrapper
    return fctn(executable, workdir, job)


# def use_payload_container(job):
#     pass


def use_middleware_container():
    """
    Should middleware from container be used?
    In case middleware, i.e. the copy command for stage-in/out, should be taken from a container this function should
    return True.

    :return: True if middleware should be taken from container. False otherwise.
    """

    if get_middleware_type() == 'container':
        return True
    else:
        return False


def get_middleware_container():
    pass


def extract_platform_and_os(platform):
    """
    Extract the platform and OS substring from platform

    :param platform (string): E.g. "x86_64-slc6-gcc48-opt"
    :return: extracted platform specifics (string). E.g. "x86_64-slc6". In case of failure, return the full platform
    """

    pattern = r"([A-Za-z0-9_-]+)-.+-.+"
    a = re.findall(re.compile(pattern), platform)

    if len(a) > 0:
        ret = a[0]
    else:
        logger.warning("could not extract architecture and OS substring using pattern=%s from platform=%s"
                       "(will use %s for image name)" % (pattern, platform, platform))
        ret = platform

    return ret


def get_grid_image_for_singularity(platform):
    """
    Return the full path to the singularity grid image

    :param platform (string): E.g. "x86_64-slc6"
    :return: full path to grid image (string).
    """

    if not platform or platform == "":
        platform = "x86_64-slc6"
        logger.warning("using default platform=%s (cmtconfig not set)" % (platform))

    arch_and_os = extract_platform_and_os(platform)
    image = arch_and_os + ".img"
    _path = os.path.join(get_file_system_root_path(), "atlas.cern.ch/repo/containers/images/singularity")
    path = os.path.join(_path, image)
    if not os.path.exists(path):
        image = 'x86_64-centos7.img'
        logger.warning('path does not exist: %s (trying with image %s instead)' % (path, image))
        path = os.path.join(_path, image)
        if not os.path.exists(path):
            logger.warning('path does not exist either: %s' % path)
            path = ""

    return path


def get_middleware_type():
    """
    Return the middleware type from the container type.
    E.g. container_type = 'singularity:pilot;docker:wrapper;middleware:container'
    get_middleware_type() -> 'container', meaning that middleware should be taken from the container. The default
    is otherwise 'workernode', i.e. middleware is assumed to be present on the worker node.

    :return: middleware_type (string)
    """

    middleware_type = ""
    container_type = infosys.queuedata.container_type

    mw = 'middleware'
    if container_type and container_type != "" and mw in container_type:
        try:
            container_names = container_type.split(';')
            for name in container_names:
                t = name.split(':')
                if mw == t[0]:
                    middleware_type = t[1]
        except Exception as e:
            logger.warning("failed to parse the container name: %s, %s" % (container_type, e))
    else:
        # logger.warning("container middleware type not specified in queuedata")
        # no middleware type was specified, assume that middleware is present on worker node
        middleware_type = "workernode"

    return middleware_type


def alrb_wrapper(cmd, workdir, job):
    """
    Wrap the given command with the special ALRB setup for containers
    E.g. cmd = /bin/bash hello_world.sh
    ->
    export thePlatform="x86_64-slc6-gcc48-opt"
    export ALRB_CONT_RUNPAYLOAD="cmd'
    setupATLAS -c $thePlatform

    :param cmd (string): command to be executed in a container.
    :param workdir: (not used)
    :param job: job object.
    :return: prepended command with singularity execution command (string).
    """

    log = get_logger(job.jobid)
    queuedata = job.infosys.queuedata

    container_name = queuedata.container_type.get("pilot")  # resolve container name for user=pilot
    if container_name == 'singularity':
        # first get the full setup, which should be removed from cmd (or ALRB setup won't work)
        _asetup = get_asetup()
        cmd = cmd.replace(_asetup, "asetup ")
        # get simplified ALRB setup (export)
        asetup = get_asetup(alrb=True)

        # Get the singularity options
        singularity_options = queuedata.container_options
        log.debug(
            "resolved singularity_options from queuedata.container_options: %s" % singularity_options)

        _cmd = asetup

        # do not include the X509_USER_PROXY in the command the container will execute
        x509 = os.environ.get('X509_USER_PROXY')
        if x509 != "":
            cmd = cmd.replace("export X509_USER_PROXY=%s;" % x509, "")
            # add it instead to the container setup command
            _cmd = "export X509_USER_PROXY=%s;" % x509 + _cmd

        if job.alrbuserplatform:
            _cmd += 'export thePlatform=\"%s\";' % job.alrbuserplatform
        elif job.platform:
            _cmd += 'export thePlatform=\"%s\";' % job.platform
        #elif '--containerImage' in job.jobparams:
        #    if job.alrbuserplatform:
        #        _cmd += 'export thePlatform=\"%s\";' % job.alrbuserplatform
        #    else:
        #        # set a default platform for user defined containers
        #        _cmd += 'export thePlatform=\"centos7\";'

        #if '--containall' not in singularity_options:
        #    singularity_options += ' --containall'
        if singularity_options != "":
            _cmd += 'export ALRB_CONT_CMDOPTS=\"%s\";' % singularity_options
        else:
            _cmd += 'export ALRB_CONT_CMDOPTS=\"$ALRB_CONT_CMDOPTS -C\";'

        # add the jobid to be used as an identifier for the payload running inside the container
        _cmd += "export PANDAID=%s;" % job.jobid

        # add TMPDIR
        cmd = "export TMPDIR=/srv;" + cmd

        # write the full payload command to a script file
        script_file = config.Container.script_file
        status = write_file(os.path.join(job.workdir, script_file), cmd, mute=False)
        if status:
            script_cmd = '. /srv/' + script_file
            _cmd += "export ALRB_CONT_RUNPAYLOAD=\'%s\';" % script_cmd
        else:
            log.warning('attempting to quote command instead')
            _cmd += 'export ALRB_CONT_RUNPAYLOAD=%s;' % pipes.quote(cmd)

        # also store the command string in the job object
        job.command = cmd

        # this should not be necessary after the extract_container_image() in JobData update
        # containerImage should have been removed already
        if '--containerImage' in job.jobparams:
            job.jobparams, container_path = remove_container_string(job.jobparams)
            if job.alrbuserplatform:
                _cmd += 'source ${ATLAS_LOCAL_ROOT_BASE}/user/atlasLocalSetup.sh -c %s' % job.alrbuserplatform
            elif container_path != "":
                _cmd += 'source ${ATLAS_LOCAL_ROOT_BASE}/user/atlasLocalSetup.sh -c %s' % container_path

                #if not job.platform:
                #    # add the default platform if not set by the job definition
                #    _cmd += ' $thePlatform'
            else:
                log.warning('failed to extract container path from %s' % job.jobparams)
                _cmd = ""
        else:
            # _cmd += 'source ${ATLAS_LOCAL_ROOT_BASE}/user/atlasLocalSetup.sh -c images'
            _cmd += 'source ${ATLAS_LOCAL_ROOT_BASE}/user/atlasLocalSetup.sh '
            if job.platform or job.alrbuserplatform:
                # _cmd += '+$thePlatform'
                _cmd += '-c $thePlatform'

        _cmd = _cmd.replace('  ', ' ')
        cmd = _cmd

        log.info("Updated command: %s" % cmd)
    else:
        log.warning('container %s not supported' % container_name)

    return cmd


## DEPRECATED, remove after verification with user container job
def remove_container_string(job_params):
    """ Retrieve the container string from the job parameters """

    pattern = r" \'?\-\-containerImage\=?\ ?([\S]+)\ ?\'?"
    compiled_pattern = re.compile(pattern)

    # remove any present ' around the option as well
    job_params = re.sub(r'\'\ \'', ' ', job_params)

    # extract the container path
    found = re.findall(compiled_pattern, job_params)
    container_path = found[0] if len(found) > 0 else ""

    # Remove the pattern and update the job parameters
    job_params = re.sub(pattern, ' ', job_params)

    return job_params, container_path


def singularity_wrapper(cmd, workdir, job):
    """
    Prepend the given command with the singularity execution command
    E.g. cmd = /bin/bash hello_world.sh
    -> singularity_command = singularity exec -B <bindmountsfromcatchall> <img> /bin/bash hello_world.sh
    singularity exec -B <bindmountsfromcatchall>  /cvmfs/atlas.cern.ch/repo/images/singularity/x86_64-slc6.img <script>

    :param cmd (string): command to be prepended.
    :param workdir: explicit work directory where the command should be executed (needs to be set for Singularity).
    :param job: job object.
    :return: prepended command with singularity execution command (string).
    """

    log = get_logger(job.jobid)
    queuedata = job.infosys.queuedata

    container_name = queuedata.container_type.get("pilot")  # resolve container name for user=pilot
    log.debug("resolved container_name from queuedata.contaner_type: %s" % container_name)

    if container_name == 'singularity':
        log.info("singularity has been requested")

        # Get the singularity options
        singularity_options = queuedata.container_options
        if singularity_options != "":
            singularity_options += ","
        else:
            singularity_options = "-B "
        singularity_options += "/cvmfs,${workdir},/home"
        log.debug("using singularity_options: %s" % singularity_options)

        # Get the image path
        image_path = job.imagename or get_grid_image_for_singularity(job.platform)

        # Does the image exist?
        if image_path:
            # Prepend it to the given command
            cmd = "export workdir=" + workdir + "; singularity --verbose exec " + singularity_options + " " + image_path + \
                  " /bin/bash -c " + pipes.quote("cd $workdir;pwd;%s" % cmd)

            # for testing user containers
            # singularity_options = "-B $PWD:/data --pwd / "
            # singularity_cmd = "singularity exec " + singularity_options + image_path
            # cmd = re.sub(r'-p "([A-Za-z0-9.%/]+)"', r'-p "%s\1"' % urllib.pathname2url(singularity_cmd), cmd)
        else:
            log.warning("singularity options found but image does not exist")

        log.info("updated command: %s" % cmd)

    return cmd



#!/usr/bin/env python
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Paul Nilsson, paul.nilsson@cern.ch, 2018

# from pilot.util.container import execute

import os
import logging

from pilot.user.atlas.setup import get_file_system_root_path
from pilot.util.container import execute
from pilot.common.errorcodes import ErrorCodes

logger = logging.getLogger(__name__)
errors = ErrorCodes()


def verify_proxy(limit=None):
    """
    Check for a valid voms/grid proxy longer than N hours.
    Use `limit` to set required time limit.

    :param limit: time limit in hours (int).
    :return: exit code (NOPROXY or NOVOMSPROXY), diagnostics (error diagnostics string).
    """

    exit_code = 0
    diagnostics = ""

    if limit is None:
        limit = 48

    # add setup for arcproxy if it exists
    #arcproxy_setup = "%s/atlas.cern.ch/repo/sw/arc/client/latest/slc6/x86_64/setup.sh" % get_file_system_root_path()
    x509 = os.environ.get('X509_USER_PROXY', '')
    if x509 != '':
        envsetup = 'export X509_USER_PROXY=%s;' % x509
    else:
        envsetup = ''
    #envsetup += ". %s;" % (arcproxy_setup)
    envsetup += ". %s/atlas.cern.ch/repo/ATLASLocalRootBase/user/atlasLocalSetup.sh --quiet;" % get_file_system_root_path()
    if os.environ.get('ALRB_noGridMW', '').lower() != "yes":
        envsetup += "lsetup emi;"
    else:
        logger.warning('Skipping "lsetup emi" as ALRB_noGridMW=YES')

    # first try to use arcproxy since voms-proxy-info is not working properly on SL6
    #  (memory issues on queues with limited memory)

    ec, diagnostics = verify_arcproxy(envsetup, limit)
    if ec == 0:
        return ec, diagnostics

    ec, diagnostics = verify_vomsproxy(envsetup, limit)
    if ec == 0:
        return ec, diagnostics

    ec, diagnostics = verify_gridproxy(envsetup, limit)
    if ec == 0:
        return ec, diagnostics

    logger.warning('none of the proxy verification methods worked - skipping verification')

    return exit_code, diagnostics


def verify_arcproxy(envsetup, limit):
    """
    Verify the proxy using arcproxy.

    :param envsetup: general setup string for proxy commands (string).
    :param limit: time limit in hours (int).
    :return: exit code (int), error diagnostics (string).
    """

    ec = 0
    diagnostics = ""

    cmd = "%sarcproxy -i vomsACvalidityLeft" % (envsetup)

    exit_code, stdout, stderr = execute(cmd, shell=True)
    if stdout is not None:
        if "command not found" in stdout:
            logger.warning("arcproxy is not available on this queue,"
                           "this can lead to memory issues with voms-proxy-info on SL6: %s" % (stdout))
        else:
            ec, diagnostics = interpret_proxy_info(exit_code, stdout, stderr, limit)
            if ec == 0:
                logger.info("voms proxy verified using arcproxy")
                return 0, diagnostics
            elif ec == errors.NOVOMSPROXY:
                return ec, diagnostics
            else:
                logger.info("will try voms-proxy-info instead")
    else:
        logger.warning('command execution failed')

    return ec, diagnostics


def verify_vomsproxy(envsetup, limit):
    """
    Verify proxy using voms-proxy-info command.

    :param envsetup: general setup string for proxy commands (string).
    :param limit: time limit in hours (int).
    :return: exit code (int), error diagnostics (string).
    """

    ec = 0
    diagnostics = ""

    if os.environ.get('X509_USER_PROXY', '') != '':
        cmd = "%svoms-proxy-info -actimeleft --file $X509_USER_PROXY" % (envsetup)
        logger.info('executing command: %s' % cmd)
        exit_code, stdout, stderr = execute(cmd, shell=True)
        if stdout is not None:
            if "command not found" in stdout:
                logger.info("skipping voms proxy check since command is not available")
            else:
                ec, diagnostics = interpret_proxy_info(exit_code, stdout, stderr, limit)
                if ec == 0:
                    logger.info("voms proxy verified using voms-proxy-info")
                    return 0, diagnostics
        else:
            logger.warning('command execution failed')
    else:
        logger.warning('X509_USER_PROXY is not set')

    return ec, diagnostics


def verify_gridproxy(envsetup, limit):
    """
    Verify proxy using grid-proxy-info command.

    :param envsetup: general setup string for proxy commands (string).
    :param limit: time limit in hours (int).
    :return: exit code (int), error diagnostics (string).
    """

    ec = 0
    diagnostics = ""

    if limit:
        # next clause had problems: grid-proxy-info -exists -valid 0.166666666667:00
        #cmd = "%sgrid-proxy-info -exists -valid %s:00" % (envsetup, str(limit))
        # more accurate calculation of HH:MM
        limit_hours = int(limit * 60) / 60
        limit_minutes = int(limit * 60 + .999) - limit_hours * 60
        cmd = "%sgrid-proxy-info -exists -valid %d:%02d" % (envsetup, limit_hours, limit_minutes)
    else:
        cmd = "%sgrid-proxy-info -exists -valid 24:00" % (envsetup)

    logger.info('executing command: %s' % cmd)
    exit_code, stdout, stderr = execute(cmd, shell=True)
    if stdout is not None:
        if exit_code != 0:
            if stdout.find("command not found") > 0:
                logger.info("skipping grid proxy check since command is not available")
            else:
                # Analyze exit code / stdout
                diagnostics = "grid proxy certificate does not exist or is too short: %d, %s" % (exit_code, stdout)
                logger.warning(diagnostics)
                return errors.NOPROXY, diagnostics
        else:
            logger.info("grid proxy verified")
    else:
        logger.warning('command execution failed')

    return ec, diagnostics


def interpret_proxy_info(ec, stdout, stderr, limit):
    """
    Interpret the output from arcproxy or voms-proxy-info.

    :param ec: exit code from proxy command (int).
    :param stdout: stdout from proxy command (string).
    :param stderr: stderr from proxy command (string).
    :param limit: time limit in hours (int).
    :return: exit code (int), diagnostics (string).
    """

    exitcode = 0
    diagnostics = ""

    logger.debug('stdout = %s' % stdout)
    logger.debug('stderr = %s' % stderr)

    if ec != 0:
        if "Unable to verify signature! Server certificate possibly not installed" in stdout:
            logger.warning("skipping voms proxy check: %s" % (stdout))
        # test for command errors
        elif "arcproxy:" in stdout:
            diagnostics = "arcproxy failed: %s" % (stdout)
            logger.warning(diagnostics)
            exitcode = errors.GENERALERROR
        else:
            # Analyze exit code / output
            diagnostics = "voms proxy certificate check failure: %d, %s" % (ec, stdout)
            logger.warning(diagnostics)
            exitcode = errors.NOVOMSPROXY
    else:
        # remove any additional print-outs if present, assume that the last line is the time left
        if "\n" in stdout:
            # also remove the last \n in case there is one
            if stdout[-1] == '\n':
                stdout = stdout[:-1]
            stdout = stdout.split('\n')[-1]

        # test for command errors
        if "arcproxy:" in stdout:
            diagnostics = "arcproxy failed: %s" % (stdout)
            logger.warning(diagnostics)
            exitcode = errors.GENERALERROR
        else:
            # on EMI-3 the time output is different (HH:MM:SS as compared to SS on EMI-2)
            if ":" in stdout:
                ftr = [3600, 60, 1]
                stdout = sum([a * b for a, b in zip(ftr, map(int, stdout.split(':')))])
            try:
                validity = int(stdout)
                if validity >= limit * 3600:
                    logger.info("voms proxy verified (%d s)" % (validity))
                else:
                    diagnostics = "voms proxy certificate does not exist or is too short (lifetime %d s)" % (validity)
                    logger.warning(diagnostics)
                    exitcode = errors.NOVOMSPROXY
            except ValueError as e:
                diagnostics = "failed to evalute command stdout: %s, stderr: %s, exc=%s" % (stdout, stderr, e)
                logger.warning(diagnostics)
                exitcode = errors.GENERALERROR

    return exitcode, diagnostics



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Wen Guan, wen.guan@cern.ch, 2017-2018

import json
import logging
import os
import Queue
import subprocess
import sys
import threading
import time

from pilot.eventservice.esprocess.eshook import ESHook
from pilot.eventservice.esprocess.esmanager import ESManager
from pilot.eventservice.esprocess.esmessage import MessageThread
from pilot.eventservice.esprocess.esprocess import ESProcess

if sys.version_info < (2, 7):
    import unittest2 as unittest
else:
    import unittest

logging.basicConfig(stream=sys.stderr, level=logging.DEBUG)


def check_env():
    """
    Function to check whether cvmfs is available.
    To be used to decide whether to skip some test functions.

    :returns True: if cvmfs is available. Otherwise False.
    """
    return os.path.exists('/cvmfs/atlas.cern.ch/repo/')


class TestESHook(ESHook):
    """
    A class implemented ESHook, to be used to test eventservice.
    """

    def __init__(self):
        """
        Init the hook class for tests: Read payload and event ranges from a file.
                                       Download evgen files which are needed to run payload.
        """
        with open('pilot/test/resource/eventservice_job.txt') as job_file:
            job = json.load(job_file)
            self.__payload = job['payload']
            self.__event_ranges = job['event_ranges']  # doesn't exit

        if check_env():
            process = subprocess.Popen('pilot/test/resource/download_test_es_evgen.sh', shell=True, stdout=subprocess.PIPE)
            process.wait()
            if process.returncode != 0:
                raise Exception('failed to download input files for es test: %s %s' % (process.communicate()))
        else:
            logging.info("No CVMFS. skip downloading files.")

        self.__injected_event_ranges = []
        self.__outputs = []

    def get_payload(self):
        """
        Get payload hook function for tests.

        :returns: dict {'executable': <cmd string>, 'output_file': <filename or without it>, 'error_file': <filename or without it>}
        """

        return self.__payload

    def get_event_ranges(self, num_ranges=1):
        """
        Get event ranges hook function for tests.

        :returns: dict of event ranges.
                  None if no available events.
        """
        ret = []
        for _ in range(num_ranges):
            if len(self.__event_ranges) > 0:
                event_range = self.__event_ranges.pop(0)
                ret.append(event_range)
                self.__injected_event_ranges.append(event_range)
        return ret

    def handle_out_message(self, message):
        """
        Handle ES output or error messages hook function for tests.

        :param message: a dict of parsed message.
                        For 'finished' event ranges, it's {'id': <id>, 'status': 'finished', 'output': <output>, 'cpu': <cpu>,
                                                           'wall': <wall>, 'message': <full message>}.
                        Fro 'failed' event ranges, it's {'id': <id>, 'status': 'failed', 'message': <full message>}.
        """

        print(message)
        self.__outputs.append(message)

    def get_injected_event_ranges(self):
        """
        Get event ranges injected to payload for test assertion.

        :returns: List of injected event ranges.
        """
        return self.__injected_event_ranges

    def get_outputs(self):
        """
        Get outputs for test assertion.

        :returns: List of outputs.
        """
        return self.__outputs


class TestESMessageThread(unittest.TestCase):
    """
    Unit tests for event service message thread.
    """

    @unittest.skipIf(not check_env(), "No CVMFS")
    def test_msg_thread(self):
        """
        Make sure that es message thread works as expected.
        """
        queue = Queue.Queue()
        msg_thread = MessageThread(queue, socket_name='test', context='local')
        self.assertIsInstance(msg_thread, threading.Thread)

        msg_thread.start()
        time.sleep(1)
        self.assertTrue(msg_thread.is_alive())

        msg_thread.send('test')
        msg_thread.stop()
        self.assertTrue(msg_thread.is_stopped())
        time.sleep(1)
        self.assertFalse(msg_thread.is_alive())


@unittest.skipIf(not check_env(), "No CVMFS")
class TestESProcess(unittest.TestCase):
    """
    Unit tests for event service process functions
    """

    @classmethod
    def setUpClass(cls):
        cls._test_hook = TestESHook()
        cls._esProcess = ESProcess(cls._test_hook.get_payload())

    def test_set_get_event_ranges_hook(self):
        """
        Make sure that no exceptions to set get_event_ranges hook.
        """

        self._esProcess.set_get_event_ranges_hook(self._test_hook.get_event_ranges)
        self.assertEqual(self._test_hook.get_event_ranges, self._esProcess.get_get_event_ranges_hook())

    def test_set_handle_out_message_hook(self):
        """
        Make sure that no exceptions to set handle_out_message hook.
        """

        self._esProcess.set_handle_out_message_hook(self._test_hook.handle_out_message)
        self.assertEqual(self._test_hook.handle_out_message, self._esProcess.get_handle_out_message_hook())

    def test_parse_out_message(self):
        """
        Make sure to parse messages from payload correctly.
        """

        output_msg = '/tmp/HITS.12164365._000300.pool.root.1.12164365-3616045203-10980024041-4138-8,ID:12164365-3616045203-10980024041-4138-8,CPU:288,WALL:303'
        ret = self._esProcess.parse_out_message(output_msg)
        self.assertEqual(ret['status'], 'finished')
        self.assertEqual(ret['id'], '12164365-3616045203-10980024041-4138-8')

        error_msg1 = 'ERR_ATHENAMP_PROCESS 130-2068634812-21368-1-4: Failed to process event range'
        ret = self._esProcess.parse_out_message(error_msg1)
        self.assertEqual(ret['status'], 'failed')
        self.assertEqual(ret['id'], '130-2068634812-21368-1-4')

        error_msg2 = "ERR_ATHENAMP_PARSE \"u'LFN': u'eta0-25.evgen.pool.root',u'eventRangeID': u'130-2068634812-21368-1-4', u'startEvent': 5\": Wrong format"
        ret = self._esProcess.parse_out_message(error_msg2)
        self.assertEqual(ret['status'], 'failed')
        self.assertEqual(ret['id'], '130-2068634812-21368-1-4')


class TestEventService(unittest.TestCase):
    """
    Unit tests for event service functions.
    """

    @unittest.skipIf(not check_env(), "No CVMFS")
    def test_init_esmanager(self):
        """
        Make sure that no exceptions to init ESManager
        """
        test_hook = TestESHook()
        es_manager = ESManager(test_hook)
        self.assertIsInstance(es_manager, ESManager)

    @unittest.skipIf(not check_env(), "No CVMFS")
    def test_run_es(self):
        """
        Make sure that ES produced all events that injected.
        """
        test_hook = TestESHook()
        es_manager = ESManager(test_hook)
        es_manager.run()
        injected_event = test_hook.get_injected_event_ranges()
        outputs = test_hook.get_outputs()

        self.assertEqual(len(injected_event), len(outputs))
        self.assertNotEqual(len(outputs), 0)



#!/usr/bin/env python
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Mario Lassnig, mario.lassnig@cern.ch, 2017
# - Daniel Drizhuk, d.drizhuk@gmail.com, 2017
# - Paul Nilsson, paul.nilsson@cern.ch, 2017


# This is a stub implementation of the information component. It retrieves
# sites, storages, and queues from AGIS and tries to locally cache them.
# No cache update is involved, just remove the .cache files.

## THIS FILE CAN BE COMPLETELY REMOVED

import collections
import hashlib
import json
import os
import urllib2
from datetime import datetime, timedelta

from pilot.util.config import config
from pilot.util.filehandling import write_json, read_json
# from pilot.common.exception import FileHandlingFailure
from pilot.util.timer import timeout

import logging
logger = logging.getLogger(__name__)


def set_location(args, site=None):  ## TO BE DEPRECATED -- not used
    """
    Set up all necessary site information.
    Resolve everything from the specified queue name, and fill extra lookup structure.

    If site is specified, return the site and storage information only.

    :param args:
    :param site:
    :return:
    """

    args.location = collections.namedtuple('location', ['queue', 'site', 'storages', 'queuedata',
                                                        'queue_info', 'site_info', 'storages_info'])

    if site is None:
        # verify that the queue is active
        url = config.Information.queues
        if url == "":
            logger.fatal('AGIS URL for queues not set')
            return False
        all_queues = retrieve_json(url)
        if args.queue not in [queue['name'] for queue in all_queues]:
            logger.critical('specified queue NOT FOUND: %s -- aborting' % args.queue)
            return False
        if not [queue for queue in all_queues if queue['name'] == args.queue and queue['state'] == 'ACTIVE']:
            logger.critical('specified queue is NOT ACTIVE: %s -- aborting' % args.queue)
            return False
        args.location.queue = str(args.queue)
        args.location.queue_info = [queue for queue in all_queues if queue['name'] == args.queue][0]

        # find the associated site
        url = config.Information.sites
        if url == "":
            logger.fatal('AGIS URL for sites not set')
            return False
        all_sites = retrieve_json(url)
        matching_sites = [queue for queue in all_queues if queue['name'] == args.queue]
        if len(matching_sites) != 1:
            logger.critical('queue is not explicitly mapped to a single site, found: %s' %
                            [tmp_site['name'] for tmp_site in matching_sites])
            return False
        else:
            args.location.site = str(matching_sites[0]['site'])
            args.location.site_info = [tmp_site for tmp_site in all_sites if tmp_site['name'] == args.location.site][0]

    else:
        # find the associated site
        url = config.Information.sites
        if url == "":
            logger.fatal('AGIS URL for sites not set')
            return False
        all_sites = retrieve_json(url)
        result = [tmp_site for tmp_site in all_sites if tmp_site['name'] == site]
        if len(result) == 0:
            raise Exception('Specified site not found: %s' % site)
        args.location.site = site
        args.location.site_info = result[0]

    # find all enabled storages at site
    url = config.Information.storages
    if url == "":
        logger.fatal('AGIS URL for storages not set')
        return False
    all_storages = retrieve_json(url)
    args.location.storages = [str(storage['name']) for storage in all_storages
                              if storage['site'] == args.location.site and storage['state'] == 'ACTIVE']
    args.location.storages_info = {}
    for tmp_storage in args.location.storages:
        args.location.storages_info[tmp_storage] = [storage for storage in all_storages
                                                    if storage['name'] == tmp_storage and
                                                    storage['state'] == 'ACTIVE'][0]

    # find the schedconfig queue data
    args.location.queuedata = get_schedconfig_queuedata(args.location.queue)

    logger.info('queue: %s' % args.location.queue)
    logger.info('site: %s' % args.location.site)
    logger.info('storages: %s' % args.location.storages)
    logger.info('queuedata: %s' % args.location.queuedata)

    return True


def get_schedconfig_queuedata(queue):
    """
    Return and store the schedconfig queuedata.

    :param queue: PanDA queue name (e.g. BNL_PROD_MCORE)
    :return: schedconfig queuedata json dictionary
    """

    # read it locally if the queuedata file already exists
    filename = os.path.join(os.environ.get('PILOT_HOME'), config.Information.queuedata)
    if os.path.exists(filename):
        queuedata = read_json(filename)
        return queuedata

    url = config.Information.schedconfig
    if url == "":
        logger.fatal('URL for schedconfig not set')
        return False
    else:
        # add the queuename to the URL
        if not url.endswith('/'):
            url += '/'
        url += queue + '.all.json'
    queuedata = retrieve_json(url)

    # also write the queuedata to disk
    if not write_json(filename, queuedata):
        logger.warning("failed to write queuedata json to file")
    else:
        logger.info("wrote queuedata to local file %s" % filename)

    return queuedata


@timeout(seconds=120)
def retrieve_json(url):
    """
    Retrieve JSON from an URL or from a local cache.
    Store the JSON in a local cache if it doesn't exist already.

    :param url:
    :return: JSON dictionary.
    """

    logger.info('retrieving: %s' % url)
    j = _read_cache(url)
    if j is not None:
        logger.debug('cached version found: %s' % url)
        return j
    j = json.loads(urllib2.urlopen(url).read())
    logger.info('caching: %s' % url)
    _write_cache(url, j)

    return j


def _read_cache(url):
    """
    Read the local cache file that contains the previously downloaded JSON from URL.

    :param url:
    :return: JSON dictionary.
    """

    j = None
    m = hashlib.md5()
    m.update(url)
    if os.path.exists('.cache.%s' % m.hexdigest()):
        with open('.cache.%s' % m.hexdigest(), 'rb') as infile:
            j = json.load(infile)

    return j


def _write_cache(url, j):
    """
    Write the JSON from URL into a local cache.

    :param url:
    :param j: JSON dictionary.
    :return:
    """

    m = hashlib.md5()
    m.update(url)
    with open('.cache.%s' % m.hexdigest(), 'wb') as outfile:
        json.dump(j, outfile)


def is_file_expired(fname, cache_time=0):
    """
    Check if file fname is older than cache_time seconds from its last_update_time.

    :param fname: File name.
    :param cache_time: Cache time in seconds.
    :return: Boolean.
    """

    if cache_time:
        lastupdate = get_file_last_update_time(fname)
        return not (lastupdate and datetime.now() - lastupdate < timedelta(seconds=cache_time))

    return True


def get_file_last_update_time(fname):
    """
    Return the last update time of the given file.

    :param fname: File name.
    :return: Last update time in seconds.
    """

    try:
        lastupdate = datetime.fromtimestamp(os.stat(fname).st_mtime)
    except Exception:
        lastupdate = None

    return lastupdate


def load_url_data(url, fname=None, cache_time=0, nretry=3, sleeptime=60):
    """
    Download data from url/file resource and optionally save it into cache file fname.
    The file will not be (re-)loaded again if cache age from last file modification does not exceed cache_time
    seconds.

    :param url:
    :param fname: File name.
    :param cache_time: Cache time in seconds.
    :param nretry: Number of retries (default is 3).
    :param sleeptime: Sleep time (default is 60 s) between retry attempts.
    :return: data loaded from the url or file content if url passed is a filename.
    """

    content = None
    if url and is_file_expired(fname, cache_time):  # load data into temporary cache file
        for trial in range(nretry):
            if content:
                break
            try:
                if os.path.isfile(url):
                    logger.info('[attempt=%s] loading data from file=%s' % (trial, url))
                    with open(url, "r") as f:
                        content = f.read()
                else:
                    logger.info('[attempt=%s] loading data from url=%s' % (trial, url))
                    content = urllib2.urlopen(url, timeout=20).read()

                if fname:  # save to cache
                    with open(fname, "w+") as f:
                        f.write(content)
                        logger.info('saved data from "%s" resource into file=%s, length=%.1fKb' %
                                    (url, fname, len(content) / 1024.))
                return content
            except Exception as e:  # ignore errors, try to use old cache if any
                logger.warning('failed to load data from url=%s, error: %s .. trying to use data from cache=%s' %
                               (url, e, fname))
                # will try to use old cache below
                if trial < nretry - 1:
                    logger.info(" -- DEPRECATED-- will try again after %ss.." % sleeptime)
                    from time import sleep
                    sleep(sleeptime)

    if content is not None:  # just loaded
        return content

    try:
        with open(fname, 'r') as f:
            content = f.read()
    except Exception as e:
        logger.warning("%s (will try different source)" % e)
        return None

    return content


def load_ddm_conf_data(ddmendpoints=[], cache_time=60):
    """
    Load DDM configuration data from a source.
    Valid sources: CVMFS, AGIS, LOCAL.

    :param ddmendpoints:
    :param cache_time: Cache time in seconds.
    :return:
    """

    # list of sources to fetch ddmconf data from
    ddmconf_sources = {'CVMFS': {'url': '/cvmfs/atlas.cern.ch/repo/sw/local/etc/agis_ddmendpoints.json',
                                 'nretry': 1,
                                 'fname': os.path.join(os.environ.get('PILOT_HOME', '.'), 'agis_ddmendpoints.cvmfs.json')},
                       'AGIS': {'url': 'http://atlas-agis-api.cern.ch/request/ddmendpoint/query/list/?json&'
                                       'state=ACTIVE&preset=dict&ddmendpoint=%s' % ','.join(ddmendpoints),
                                       'nretry': 3,
                                       'fname': os.path.join(os.environ.get('PILOT_HOME', '.'), 'agis_ddmendpoints.agis.%s.json' %
                                                             ('_'.join(sorted(ddmendpoints)) or 'ALL'))},

                       'LOCAL': {'url': None,
                                 'nretry': 1,
                                 'fname': os.path.join(os.environ.get('PILOT_HOME', '.'), 'agis_ddmendpoints.json')},
                       'PANDA': None
                       }

    ddmconf_sources_order = ['LOCAL', 'CVMFS', 'AGIS']

    for key in ddmconf_sources_order:
        logger.info("loading DDMConfData from source %s" % key)
        dat = ddmconf_sources.get(key)
        if not dat:
            continue

        content = load_url_data(cache_time=cache_time, **dat)
        if not content:
            continue
        try:
            data = json.loads(content)
        except Exception as e:
            logger.fatal("failed to parse JSON content from source=%s .. skipped, error=%s" % (dat.get('url'), e))
            data = None

        if data:
            return data

    return None


def resolve_ddm_conf(ddmendpoints):
    """
    Resolve the DDM configuration.

    :return: DDM configuration data from a source (JSON dictionary).
    """

    return load_ddm_conf_data(ddmendpoints, cache_time=6000) or {}


def load_schedconfig_data(pandaqueues=[], cache_time=60):
    """
    Download the queuedata from various sources (prioritized).
    Try to get data from CVMFS first, then AGIS or from Panda JSON sources (not implemented).
    Note: as of October 2016, agis_schedconfig.json is complete but contains info from all PQ (5 MB)

    :param pandaqueues:
    :param cache_time: Cache time in seconds.
    :return:
    """

    # list of sources to fetch ddmconf data from
    pandaqueues = set(pandaqueues)

    schedcond_sources = {'CVMFS': {'url': '/cvmfs/atlas.cern.ch/repo/sw/local/etc/agis_schedconf.json',
                                   'nretry': 1,
                                   'fname': os.path.join(os.environ.get('PILOT_HOME', '.'), 'agis_schedconf.cvmfs.json')},
                         'AGIS': {'url': 'http://atlas-agis-api.cern.ch/request/pandaqueue/query/list/?json'
                                         '&preset=schedconf.all&panda_queue=%s' % ','.join(pandaqueues),
                                  'nretry': 3,
                                  'fname': os.path.join(os.environ.get('PILOT_HOME', '.'), 'agis_schedconf.agis.%s.json' %
                                                        ('_'.join(sorted(pandaqueues)) or 'ALL'))},
                         'LOCAL': {'url': None,
                                   'nretry': 1,
                                   'fname': os.path.join(os.environ.get('PILOT_HOME', '.'), 'agis_schedconf.json')},
                         'PANDA': None}

    schedcond_sources_order = ['LOCAL', 'CVMFS', 'AGIS']

    for key in schedcond_sources_order:
        dat = schedcond_sources.get(key)
        if not dat:
            continue

        content = None
        if key == 'LOCAL':
            if os.path.exists(dat['fname']):
                with open(dat['fname']) as f:
                    content = f.read()
        else:
            content = load_url_data(cache_time=cache_time, **dat)
        if not content:
            continue
        try:
            data = json.loads(content)
        except Exception as e:
            logger.fatal("failed to parse JSON content from source=%s .. skipped, error=%s" % (dat.get('url'), e))
            data = None
        if data:
            if 'error' in data:
                logger.warning("skipped source=%s since response contains error: data=%s" % (dat.get('url'), data))
            else:
                return data

    return None


def resolve_panda_protocols(pandaqueues, activity):
    """
    Resolve PanDA protocols. WARNING: deprecated? aprotocols is always {} (in AGIS).

    Resolve (SE endpoint, path, copytool, copyprefix) protocol entry for requested ddmendpoint by given pilot activity
    ("pr" means pilot_read, "pw" for pilot_write).
    Return the list of possible protocols ordered by priority.

    :param pandaqueues: list of panda queues
    :param activity: activity (string)
    :return: dict('ddmendpoint_name':[(SE_1, path2, copytool, copyprefix), ).
    """

    if not pandaqueues:
        return {}

    schedconf = load_schedconfig_data(pandaqueues, cache_time=6000) or {}

    ret = {}
    for pandaqueue in set(pandaqueues):
        qdata = schedconf.get(pandaqueue, {})
        protocols = qdata.get('aprotocols', {}).get(activity, [])
        copytools = qdata.get('copytools', {})
        for p in protocols:
            p.setdefault('copysetup', copytools.get(p.get('copytool'), {}).get('setup'))
        ret.setdefault(pandaqueue, protocols)

    return ret


def resolve_panda_copytools(pandaqueues, activity, defval=[]):
    """
    Resolve supported copytools by given pandaqueues
    Check first settings for requested activity (pr, pw, pl, pls), then defval values,
    if not set then return copytools explicitly defined for all activities (not restricted to specific activity).
    Return ordered list of accepted copytools.

    :param pandaqueues:
    :param activity: activity of prioritized list of activities to resolve data.
    :param defval: default copytools values which will be used if no copytools defined for requested activity.
    :return: dict('pandaqueue':[(copytool, {settings}), ('copytool_name', {'setup':''}), ]).
    """

    r = load_schedconfig_data(pandaqueues, cache_time=6000) or {}

    ret = {}
    for pandaqueue in set(pandaqueues):
        copytools = r.get(pandaqueue, {}).get('copytools', {})
        cptools = []
        acopytools = None
        for a in activity:
            acopytools = r.get(pandaqueue, {}).get('acopytools', {}).get(a, [])
            if acopytools:
                break
        if acopytools:
            cptools = [(cp, copytools[cp]) for cp in acopytools if cp in copytools]
        elif defval:
            cptools = defval[:]
        else:
            explicit_copytools = set()
            for v in r.get(pandaqueue, {}).get('acopytools', {}).itervalues():
                explicit_copytools.update(v or [])

            cptools = [(cp, v) for cp, v in copytools.iteritems() if cp not in explicit_copytools]

        ret.setdefault(pandaqueue, cptools)

    return ret


def resolve_panda_os_ddms(pandaqueues):
    """
    Resolve OS ddmendpoints associated with requested pandaqueues.

    :param pandaqueues:
    :return: list of accepted ddmendpoints.
    """

    r = load_schedconfig_data(pandaqueues, cache_time=6000) or {}

    ret = {}
    for pandaqueue in set(pandaqueues):
        ret[pandaqueue] = r.get(pandaqueue, {}).get('ddmendpoints', [])

    return ret


def resolve_panda_associated_storages_for_activity(pandaqueue, activity):
    """
    Resolve DDM storages associated to requested pandaqueue for given activity

    :param pandaqueue: panda queues (string).
    :param activity: activity (string). E.g. 'pr'.
    :return:
    """

    return resolve_panda_associated_storages([pandaqueue]).get(pandaqueue, {})[activity]


def resolve_panda_associated_storages(pandaqueues):
    """
    Resolve DDM storages associated to requested pandaqueues

    :param pandaqueues: panda queues (list).
    :return: list of accepted ddmendpoints.

    :return:
    """

    r = load_schedconfig_data(pandaqueues, cache_time=6000) or {}

    ret = {}
    for pandaqueue in set(pandaqueues):
        ret[pandaqueue] = r.get(pandaqueue, {}).get('astorages', {})

    return ret


def resolve_items(pandaqueues, itemname):
    """
    Resolve DDM storages associated with requested pandaqueues.

    :param pandaqueues:
    :param itemname:
    :return: list of accepted ddmendpoints.
    """

    r = load_schedconfig_data(pandaqueues, cache_time=6000) or {}

    ret = {}
    for pandaqueue in set(pandaqueues):
        ret[pandaqueue] = r.get(pandaqueue, {}).get(itemname, {})

    return ret



#!/usr/bin/env python
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Paul Nilsson, paul.nilsson@cern.ch, 2017-2018
# - Wen Guan, wen.guan, 2018

import os
import re
import fnmatch
from collections import defaultdict
from glob import glob
from signal import SIGTERM, SIGUSR1

from .dbrelease import get_dbrelease_version, create_dbrelease
from .setup import should_pilot_prepare_asetup, get_asetup, get_asetup_options, is_standard_atlas_job,\
    set_inds, get_analysis_trf, get_payload_environment_variables, replace_lfns_with_turls
from .utilities import get_memory_monitor_setup, get_network_monitor_setup, post_memory_monitor_action,\
    get_memory_monitor_summary_filename, get_prefetcher_setup, get_benchmark_setup

from pilot.common.errorcodes import ErrorCodes
from pilot.common.exception import TrfDownloadFailure, PilotException
from pilot.util.auxiliary import get_logger
from pilot.util.config import config
from pilot.util.constants import UTILITY_BEFORE_PAYLOAD, UTILITY_WITH_PAYLOAD, UTILITY_AFTER_PAYLOAD_STARTED,\
    UTILITY_WITH_STAGEIN
from pilot.util.container import execute
from pilot.util.filehandling import remove, get_guid, remove_dir_tree, read_list

from pilot.info import FileSpec

import logging
logger = logging.getLogger(__name__)

errors = ErrorCodes()


def sanity_check():
    """
    Perform an initial sanity check before doing anything else in a given workflow.
    This function can be used to verify importing of modules that are otherwise used much later, but it is better to abort
    the pilot if a problem is discovered early.

    :return: exit code (0 if all is ok, otherwise non-zero exit code).
    """

    exit_code = 0

    #try:
    #    from rucio.client.downloadclient import DownloadClient
    #    from rucio.client.uploadclient import UploadClient
    #    # note: must do something with Download/UploadClients or flake8 will complain - but do not instantiate
    #except Exception as e:
    #    logger.warning('sanity check failed: %s' % e)
    #    exit_code = errors.MIDDLEWAREIMPORTFAILURE

    return exit_code


def validate(job):
    """
    Perform user specific payload/job validation.
    This function will produce a local DBRelease file if necessary (old releases).

    :param job: job object.
    :return: Boolean (True if validation is successful).
    """

    log = get_logger(job.jobid)
    status = True

    if 'DBRelease' in job.jobparams:
        log.debug('encountered DBRelease info in job parameters - will attempt to create a local DBRelease file')
        version = get_dbrelease_version(job.jobparams)
        if version:
            status = create_dbrelease(version, job.workdir)

    # assign error in case of DBRelease handling failure
    if not status:
        job.piloterrorcodes, job.piloterrordiags = errors.add_error_code(errors.DBRELEASEFAILURE)

    # cleanup job parameters if only copy-to-scratch
    #if job.only_copy_to_scratch():
    #    log.debug('job.params=%s' % job.jobparams)
    #    if ' --usePFCTurl' in job.jobparams:
    #        log.debug('cleaning up --usePFCTurl from job parameters since all input is copy-to-scratch')
    #        job.jobparams = job.jobparams.replace(' --usePFCTurl', '')
    #    if ' --directIn' in job.jobparams:
    #        log.debug('cleaning up --directIn from job parameters since all input is copy-to-scratch')
    #        job.jobparams = job.jobparams.replace(' --directIn', '')

    return status


def get_payload_command(job):
    """
    Return the full command for execuring the payload, including the sourcing of all setup files and setting of
    environment variables.

    :param job: job object.
    :raises PilotException: TrfDownloadFailure.
    :return: command (string).
    """

    log = get_logger(job.jobid)

    # Should the pilot do the asetup or do the jobPars already contain the information?
    prepareasetup = should_pilot_prepare_asetup(job.noexecstrcnv, job.jobparams)

    # Get the platform value
    # platform = job.infosys.queuedata.platform

    # Is it a user job or not?
    userjob = job.is_analysis()

    # get the general setup command and then verify it
    cmd = get_setup_command(job, prepareasetup)
    ec, diagnostics = verify_setup_command(cmd)
    if ec != 0:
        raise PilotException(diagnostics, code=ec)

    if is_standard_atlas_job(job.swrelease):

        # Normal setup (production and user jobs)
        log.info("preparing normal production/analysis job setup command")
        cmd = get_normal_payload_command(cmd, job, prepareasetup, userjob)

    else:  # Generic, non-ATLAS specific jobs, or at least a job with undefined swRelease

        log.info("generic job (non-ATLAS specific or with undefined swRelease)")
        cmd = get_generic_payload_command(cmd, job, prepareasetup, userjob)

    # add any missing trailing ;
    if not cmd.endswith(';'):
        cmd += '; '
    log.debug('post cmd: %s' % cmd)

    # only if not using a user container
    if not job.imagename:
        site = os.environ.get('PILOT_SITENAME', '')
        variables = get_payload_environment_variables(cmd, job.jobid, job.taskid, job.attemptnr, job.processingtype, site, userjob)
        cmd = ''.join(variables) + cmd

    cmd = cmd.replace(';;', ';')

    # For direct access in prod jobs, we need to substitute the input file names with the corresponding TURLs
    # get relevant file transfer info
    use_copy_tool, use_direct_access, use_pfc_turl = get_file_transfer_info(job)
    if not userjob and use_direct_access and job.transfertype == 'direct':
        lfns, guids = job.get_lfns_and_guids()
        cmd = replace_lfns_with_turls(cmd, job.workdir, "PoolFileCatalog.xml", lfns, writetofile=job.writetofile)

    # Explicitly add the ATHENA_PROC_NUMBER (or JOB value)
    cmd = add_athena_proc_number(cmd)

    log.info('payload run command: %s' % cmd)

    return cmd


def verify_setup_command(cmd):
    """
    Verify the setup command.

    :param cmd: command string to be verified (string).
    :return: pilot error code (int), diagnostics (string).
    """

    ec = 0
    diagnostics = ""

    exit_code, stdout, stderr = execute(cmd, timeout=5 * 60)
    if exit_code != 0:
        if "No release candidates found" in stdout:
            logger.info('exit_code=%d' % exit_code)
            logger.info('stdout=%s' % stdout)
            logger.info('stderr=%s' % stderr)
            ec = errors.NORELEASEFOUND
            diagnostics = stdout + stderr

    return ec, diagnostics


def get_normal_payload_command(cmd, job, prepareasetup, userjob):
    """
    Return the payload command for a normal production/analysis job.

    :param cmd: any preliminary command setup (string).
    :param job: job object.
    :param userjob: True for user analysis jobs, False otherwise (bool).
    :param prepareasetup: True if the pilot should prepare the setup, False if already in the job parameters.
    :return: normal payload command (string).
    """

    log = get_logger(job.jobid)

    # set the INDS env variable (used by runAthena but also for EventIndex production jobs)
    set_inds(job.datasetin)  # realDatasetsIn

    if userjob:
        # Try to download the trf (skip when user container is to be used)
        #if job.imagename != "" or "--containerImage" in job.jobparams:
        #    job.transformation = os.path.join(os.path.dirname(job.transformation), "runcontainer")
        #    log.warning('overwrote job.transformation, now set to: %s' % job.transformation)
        ec, diagnostics, trf_name = get_analysis_trf(job.transformation, job.workdir)
        if ec != 0:
            raise TrfDownloadFailure(diagnostics)
        else:
            log.debug('user analysis trf: %s' % trf_name)

        if prepareasetup:
            _cmd = get_analysis_run_command(job, trf_name)
        else:
            _cmd = job.jobparams

        log.debug('job imagename: %s' % job.imagename)
        if job.imagename == "":
            # if '--containerImage' not in job.jobparams:
            # Correct for multi-core if necessary (especially important in case coreCount=1 to limit parallel make)
            cmd += "; " + add_makeflags(job.corecount, "") + _cmd
    else:
        # Add Database commands if they are set by the local site
        cmd += os.environ.get('PILOT_DB_LOCAL_SETUP_CMD', '')

        if job.is_eventservice:
            if job.corecount:
                cmd += '; export ATHENA_PROC_NUMBER=%s' % job.corecount
                cmd += '; export ATHENA_CORE_NUMBER=%s' % job.corecount
            else:
                cmd += '; export ATHENA_PROC_NUMBER=1'
                cmd += '; export ATHENA_CORE_NUMBER=1'

        # Add the transform and the job parameters (production jobs)
        if prepareasetup:
            cmd += "; %s %s" % (job.transformation, job.jobparams)
        else:
            cmd += "; " + job.jobparams

    return cmd


def get_generic_payload_command(cmd, job, prepareasetup, userjob):
    """

    :param cmd:
    :param job: job object.
    :param prepareasetup:
    :param userjob: True for user analysis jobs, False otherwise (bool).
    :return: generic job command (string).
    """

    log = get_logger(job.jobid)

    if userjob:
        # Try to download the trf
        #if job.imagename != "" or "--containerImage" in job.jobparams:
        #    job.transformation = os.path.join(os.path.dirname(job.transformation), "runcontainer")
        #    log.warning('overwrote job.transformation, now set to: %s' % job.transformation)
        ec, diagnostics, trf_name = get_analysis_trf(job.transformation, job.workdir)
        if ec != 0:
            raise TrfDownloadFailure(diagnostics)
        else:
            log.debug('user analysis trf: %s' % trf_name)

        if prepareasetup:
            _cmd = get_analysis_run_command(job, trf_name)
        else:
            _cmd = job.jobparams

        # correct for multi-core if necessary (especially important in case coreCount=1 to limit parallel make)
        # only if not using a user container
        if not job.imagename:
            cmd += "; " + add_makeflags(job.corecount, "") + _cmd
        else:
            cmd += _cmd

    elif verify_release_string(job.homepackage) != 'NULL' and job.homepackage != ' ':
        if prepareasetup:
            cmd = "python %s/%s %s" % (job.homepackage, job.transformation, job.jobparams)
        else:
            cmd = job.jobparams
    else:
        if prepareasetup:
            cmd = "python %s %s" % (job.transformation, job.jobparams)
        else:
            cmd = job.jobparams

    return cmd


def add_athena_proc_number(cmd):
    """
    Add the ATHENA_PROC_NUMBER to the payload command if necessary

    :param cmd: payload execution command (string).
    :return: updated payload execution command (string).
    """

    if "ATHENA_PROC_NUMBER" not in cmd:
        if "ATHENA_PROC_NUMBER" in os.environ:
            cmd = 'export ATHENA_PROC_NUMBER=%s;' % os.environ['ATHENA_PROC_NUMBER'] + cmd
        elif "ATHENA_PROC_NUMBER_JOB" in os.environ:
            try:
                value = int(os.environ['ATHENA_PROC_NUMBER_JOB'])
            except Exception:
                logger.warning("failed to convert ATHENA_PROC_NUMBER_JOB=%s to int" %
                               os.environ['ATHENA_PROC_NUMBER_JOB'])
            else:
                if value > 1:
                    cmd = 'export ATHENA_PROC_NUMBER=%d;' % value + cmd
                else:
                    logger.info("will not add ATHENA_PROC_NUMBER to cmd since the value is %d" % value)
        else:
            logger.warning("don't know how to set ATHENA_PROC_NUMBER (could not find it in os.environ)")
    else:
        logger.info("ATHENA_PROC_NUMBER already in job command")

    return cmd


def get_setup_command(job, prepareasetup):
    """
    Return the path to asetup command, the asetup command itself and add the options (if desired).
    If prepareasetup is False, the function will only return the path to the asetup script. It is then assumed
    to be part of the job parameters.

    :param job: job object.
    :param prepareasetup: should the pilot prepare the asetup command itself? boolean.
    :return:
    """

    # return immediately if there is no release or if user containers are used
    if job.swrelease == 'NULL' or '--containerImage' in job.jobparams:
        return ""

    # Define the setup for asetup, i.e. including full path to asetup and setting of ATLAS_LOCAL_ROOT_BASE
    cmd = get_asetup(asetup=prepareasetup)

    if prepareasetup:
        options = get_asetup_options(job.swrelease, job.homepackage)
        asetupoptions = " " + options + " --platform " + job.platform

        # Always set the --makeflags option (to prevent asetup from overwriting it)
        asetupoptions += " --makeflags=\'$MAKEFLAGS\'"

        # Verify that the setup works
        # exitcode, output = timedCommand(cmd, timeout=5 * 60)
        # if exitcode != 0:
        #     if "No release candidates found" in output:
        #         pilotErrorDiag = "No release candidates found"
        #         logger.warning(pilotErrorDiag)
        #         return self.__error.ERR_NORELEASEFOUND, pilotErrorDiag, "", special_setup_cmd, JEM, cmtconfig
        # else:
        #     logger.info("verified setup command")

        cmd += asetupoptions

    return cmd


def verify_release_string(release):
    """
    Verify that the release (or homepackage) string is set.

    :param release: release or homepackage string that might or might not be set.
    :return: release (set string).
    """

    if release is None:
        release = ""
    release = release.upper()
    if release == "":
        release = "NULL"
    if release == "NULL":
        logger.info("detected unset (NULL) release/homepackage string")

    return release


def add_makeflags(job_core_count, cmd):
    """
    Correct for multi-core if necessary (especially important in case coreCount=1 to limit parallel make).

    :param job_core_count: core count from the job definition (int).
    :param cmd: payload execution command (string).
    :return: updated payload execution command (string).
    """

    # ATHENA_PROC_NUMBER is set in Node.py using the schedconfig value
    try:
        core_count = int(os.environ.get('ATHENA_PROC_NUMBER'))
    except Exception:
        core_count = -1
    if core_count == -1:
        try:
            core_count = int(job_core_count)
        except Exception:
            pass
        else:
            if core_count >= 1:
                # Note: the original request (AF) was to use j%d and not -j%d, now using the latter
                cmd += "export MAKEFLAGS=\'-j%d QUICK=1 -l1\';" % (core_count)

    # make sure that MAKEFLAGS is always set
    if "MAKEFLAGS=" not in cmd:
        cmd += "export MAKEFLAGS=\'-j1 QUICK=1 -l1\';"

    return cmd


def get_analysis_run_command(job, trf_name):
    """
    Return the proper run command for the user job.

    Example output: export X509_USER_PROXY=<..>;./runAthena <job parameters> --usePFCTurl --directIn

    :param job: job object.
    :param trf_name: name of the transform that will run the job (string). Used when containers are not used.
    :return: command (string).
    """

    cmd = ""

    log = get_logger(job.jobid)

    # get relevant file transfer info
    use_copy_tool, use_direct_access, use_pfc_turl = get_file_transfer_info(job)
    # check if the input files are to be accessed locally (ie if prodDBlockToken is set to local)
    if job.is_local():
        log.debug('switched off direct access for local prodDBlockToken')
        use_direct_access = False
        use_pfc_turl = False

    # add the user proxy
    if 'X509_USER_PROXY' in os.environ and not job.imagename:
        cmd += 'export X509_USER_PROXY=%s;' % os.environ.get('X509_USER_PROXY')

    # set up analysis trf
    if job.imagename == "":
        cmd += './%s %s' % (trf_name, job.jobparams)
    else:
        cmd += 'python %s %s' % (trf_name, job.jobparams)

        # restore the image name
        cmd += ' --containerImage=%s' % job.imagename

    # add control options for PFC turl and direct access
    if job.indata != []:
        if use_pfc_turl and '--usePFCTurl' not in cmd:
            cmd += ' --usePFCTurl'
        if use_direct_access and '--directIn' not in cmd:
            cmd += ' --directIn'

    # update the payload command for forced accessmode
    cmd = update_forced_accessmode(log, cmd, job.transfertype, job.jobparams, trf_name)

    # add guids when needed
    # get the correct guids list (with only the direct access files)
    if not job.is_build_job():
        lfns, guids = job.get_lfns_and_guids()
        _guids = get_guids_from_jobparams(job.jobparams, lfns, guids)
        if _guids:
            cmd += ' --inputGUIDs \"%s\"' % (str(_guids))

    return cmd


def update_forced_accessmode(log, cmd, transfertype, jobparams, trf_name):
    """
    Update the payload command for forced accessmode.
    accessmode is an option that comes from HammerCloud and is used to force a certain input file access mode; i.e.
    copy-to-scratch or direct access.

    :param log: logging object.
    :param cmd: payload command.
    :param transfertype: transfer type (.e.g 'direct') from the job definition with priority over accessmode (string).
    :param jobparams: job parameters (string).
    :param trf_name: transformation name (string).
    :return: updated payload command string.
    """

    if "accessmode" in cmd and transfertype != 'direct':
        accessmode_usect = None
        accessmode_directin = None
        _accessmode_dic = {"--accessmode=copy": ["copy-to-scratch mode", ""],
                           "--accessmode=direct": ["direct access mode", " --directIn"]}

        # update run_command according to jobPars
        for _mode in _accessmode_dic.keys():
            if _mode in jobparams:
                # any accessmode set in jobPars should overrule schedconfig
                log.info("enforcing %s" % _accessmode_dic[_mode][0])
                if _mode == "--accessmode=copy":
                    # make sure direct access is turned off
                    accessmode_usect = True
                    accessmode_directin = False
                elif _mode == "--accessmode=direct":
                    # make sure copy-to-scratch gets turned off
                    accessmode_usect = False
                    accessmode_directin = True
                else:
                    accessmode_usect = False
                    accessmode_directin = False

                # update run_command (do not send the accessmode switch to runAthena)
                cmd += _accessmode_dic[_mode][1]
                if _mode in cmd:
                    cmd = cmd.replace(_mode, "")

        # force usage of copy tool for stage-in or direct access
        if accessmode_usect:
            log.info('forced copy tool usage selected')
            # remove again the "--directIn"
            if "directIn" in cmd:
                cmd = cmd.replace(' --directIn', ' ')
        elif accessmode_directin:
            log.info('forced direct access usage selected')
            if "directIn" not in cmd:
                cmd += ' --directIn'
        else:
            log.warning('neither forced copy tool usage nor direct access was selected')

        if "directIn" in cmd and "usePFCTurl" not in cmd:
            cmd += ' --usePFCTurl'

        # need to add proxy if not there already
        if "--directIn" in cmd and "export X509_USER_PROXY" not in cmd:
            if 'X509_USER_PROXY' in os.environ:
                cmd = cmd.replace("./%s" % trf_name, "export X509_USER_PROXY=%s;./%s" %
                                  (os.environ.get('X509_USER_PROXY'), trf_name))

    # if both direct access and the accessmode loop added a directIn switch, remove the first one from the string
    if cmd.count("directIn") > 1:
        cmd = cmd.replace(' --directIn', ' ', 1)

    return cmd


def get_guids_from_jobparams(jobparams, infiles, infilesguids):
    """
    Extract the correct guid from the input file list.
    The guids list is used for direct reading.
    1. extract input file list for direct reading from job parameters
    2. for each input file in this list, find the corresponding guid from the input file guid list
    Since the job parameters string is entered by a human, the order of the input files might not be the same.

    :param jobparams: job parameters.
    :param infiles: input file list.
    :param infilesguids: input file guids list.
    :return: guids list.
    """

    guidlist = []
    jobparams = jobparams.replace("'", "")
    jobparams = jobparams.replace(", ", ",")

    pattern = re.compile(r'\-i \"\[([A-Za-z0-9.,_-]+)\]\"')
    directreadinginputfiles = re.findall(pattern, jobparams)
    _infiles = []
    if directreadinginputfiles != []:
        _infiles = directreadinginputfiles[0].split(",")
    else:
        match = re.search("-i ([A-Za-z0-9.\[\],_-]+) ", jobparams)
        if match is not None:
            compactinfiles = match.group(1)
            match = re.search('(.*)\[(.+)\](.*)\[(.+)\]', compactinfiles)
            if match is not None:
                infiles = []
                head = match.group(1)
                tail = match.group(3)
                body = match.group(2).split(',')
                attr = match.group(4).split(',')
                for idx in range(len(body)):
                    lfn = '%s%s%s%s' % (head, body[idx], tail, attr[idx])
                    infiles.append(lfn)
            else:
                infiles = [compactinfiles]

    if _infiles != []:
        for infile in _infiles:
            # get the corresponding index from the inputFiles list, which has the same order as infilesguids
            try:
                index = infiles.index(infile)
            except Exception as e:
                logger.warning("exception caught: %s (direct reading will fail)" % e)
            else:
                # add the corresponding guid to the list
                guidlist.append(infilesguids[index])

    return guidlist


def get_file_transfer_info(job):
    """
    Return information about desired file transfer.

    :param job: job object
    :return: use copy tool (boolean), use direct access (boolean), use PFC Turl (boolean).
    """

    use_copy_tool = True
    use_direct_access = False
    use_pfc_turl = False

    # check with schedconfig
    if (job.infosys.queuedata.direct_access_lan or job.infosys.queuedata.direct_access_wan or job.transfertype == 'direct') and not job.is_build_job():
        # override if all input files are copy-to-scratch
        if job.only_copy_to_scratch():
            logger.info('all input files are copy-to-scratch (--usePFCTurl and --directIn will not be set)')
        else:
            logger.debug('--usePFCTurl and --directIn will be set')
            use_copy_tool = False
            use_direct_access = True
            use_pfc_turl = True

    return use_copy_tool, use_direct_access, use_pfc_turl


def update_job_data(job):
    """
    This function can be used to update/add data to the job object.
    E.g. user specific information can be extracted from other job object fields. In the case of ATLAS, information
    is extracted from the metadata field and added to other job object fields.

    :param job: job object
    :return:
    """

    ## comment from Alexey:
    ## it would be better to reallocate this logic (as well as parse metadata values)directly to Job object
    ## since in general it's Job related part
    ## later on once we introduce VO specific Job class (inherited from JobData) this can be easily customized

    log = get_logger(job.jobid)

    stageout = "all"

    if job.is_eventservice:
        logger.info('event service payload, will only stage-out log')
        stageout = "log"
    else:
        # handle any error codes
        if 'exeErrorCode' in job.metadata:
            job.exeerrorcode = job.metadata['exeErrorCode']
            if job.exeerrorcode == 0:
                stageout = "all"
            else:
                log.info('payload failed: exeErrorCode=%d' % job.exeerrorcode)
                stageout = "log"

    if 'exeErrorDiag' in job.metadata:
        job.exeerrordiag = job.metadata['exeErrorDiag']
        if job.exeerrordiag:
            log.warning('payload failed: exeErrorDiag=%s' % job.exeerrordiag)

    # determine what should be staged out
    job.stageout = stageout  # output and log file or only log file

    work_attributes = None
    try:
        work_attributes = parse_jobreport_data(job.metadata)
    except Exception as e:
        log.warning('failed to parse job report: %s' % e)

    log.info('work_attributes = %s' % work_attributes)

    # note: the number of events can be set already at this point if the value was extracted from the job report
    # (a more thorough search for this value is done later unless it was set here)
    nevents = work_attributes.get('nEvents', 0)
    if nevents:
        job.nevents = nevents

    # extract output files from the job report if required, in case the trf has created additional (overflow) files
    # also make sure all guids are assigned (use job report value if present, otherwise generate the guid)
    if job.metadata and not job.is_eventservice and not job.is_analysis():
        extract_output_files(job)  # keep this for now, complicated to merge with verify_output_files?
        verify_output_files(job)
    else:
        if not job.allownooutput:  # i.e. if it's an empty list/string, do nothing
            log.debug("will not try to extract output files from jobReport for user job (and allowNoOut list is empty)")
        else:
            # remove the files listed in allowNoOutput if they don't exist
            remove_no_output_files(job)

    ## validate output data (to be moved into the JobData)
    ## warning: do no execute this code unless guid lookup in job report has failed - pilot should only generate guids
    ## if they are not present in job report
    for dat in job.outdata:
        if not dat.guid:
            dat.guid = get_guid()
            log.warning('guid not set: generated guid=%s for lfn=%s' % (dat.guid, dat.lfn))


def extract_output_files(job):
    """
    Extract output files from the job report if required (not for user jobs), in case the trf has created additional
    (overflow) files also make sure all guids are assigned (use job report value if present, otherwise generate the guid)

    :param job: job object.
    :return:
    """

    log = get_logger(job.jobid)

    # make sure there is a defined output file list in the job report - unless it is allowed by task parameter allowNoOutput
    if not job.allownooutput:
        output = job.metadata.get('files', {}).get('output', [])
        if output:
            logger.info('verified that job report contains metadata for %d file(s)' % len(output))
        else:
            logger.warning('job report contains no output files - will fail job since allowNoOutput is not set')
            job.piloterrorcodes, job.piloterrordiags = errors.add_error_code(errors.NOOUTPUTINJOBREPORT)

    # extract info from metadata (job report JSON)
    data = dict([e.lfn, e] for e in job.outdata)
    extra = []
    for dat in job.metadata.get('files', {}).get('output', []):
        for fdat in dat.get('subFiles', []):
            lfn = fdat['name']

            # verify the guid if the lfn is known
            if lfn in data:
                data[lfn].guid = fdat['file_guid']
                logger.info('set guid=%s for lfn=%s (value taken from job report)' % (data[lfn].guid, lfn))
            else:  # found new entry, create filespec
                #if not job.outdata:
                #    raise PilotException("job.outdata is empty, will not be able to construct FileSpecs",
                #                         code=errors.INTERNALPILOTPROBLEM)
                if job.outdata:
                    kw = {'lfn': lfn,
                          'scope': job.outdata[0].scope,  ## take value from 1st output file?
                          'guid': fdat['file_guid'],
                          'filesize': fdat['file_size'],
                          'dataset': dat.get('dataset') or job.outdata[0].dataset  ## take value from 1st output file?
                          }
                    spec = FileSpec(filetype='output', **kw)
                    extra.append(spec)

    if extra:
        log.info('found extra output files in job report, will overwrite output file list: extra=%s' % extra)
        job.outdata = extra


def verify_output_files(job):  # noqa: C901
    """
    Make sure that the known output files from the job definition are listed in the job report and number of processed events
    is greater than zero. If the output file is not listed in the job report, then if the file is listed in allowNoOutput
    remove it from stage-out, otherwise fail the job.

    Note from Rod: fail scenario: The output file is not in output:[] or is there with zero events. Then if allownooutput is not
    set - fail the job. If it is set, then do not store the output, and finish ok.

    :param job: job object.
    :return: Boolean (and potentially updated job.outdata list)
    """

    failed = False
    log = get_logger(job.jobid)

    # get list of output files from the job definition
    lfns_jobdef = []
    for fspec in job.outdata:
        lfns_jobdef.append(fspec.lfn)
    if not lfns_jobdef:
        log.debug('empty output file list from job definition (nothing to verify)')
        return True

    # get list of output files from job report
    # (if None is returned, it means the job report is from an old release and does not contain an output list)
    output = job.metadata.get('files', {}).get('output', None)
    if not output and output is not None:
        # ie empty list, output=[] - are all known output files in allowNoOutput?
        log.warning('encountered an empty output file list in job report, consulting allowNoOutput list')
        failed = False
        for lfn in lfns_jobdef:
            if lfn not in job.allownooutput:
                failed = True
                log.warning('lfn %s is not in allowNoOutput list - job will fail' % lfn)
                job.piloterrorcodes, job.piloterrordiags = errors.add_error_code(errors.MISSINGOUTPUTFILE)
                break
            else:
                log.info('lfn %s listed in allowNoOutput - will be removed from stage-out' % lfn)
                remove_from_stageout(lfn, job)

    elif output is None:
        # ie job report is ancient / output could not be extracted
        log.warning('output file list could not be extracted from job report (nothing to verify)')
    else:
        output_jobrep = {}  # {lfn: nentries, ..}
        log.debug('extracted output file list from job report - make sure all known output files are listed')
        failed = False
        # first collect the output files from the job report
        for dat in output:
            for fdat in dat.get('subFiles', []):
                # get the lfn
                name = fdat.get('name', None)

                # get the number of processed events
                nentries = fdat.get('nentries', None)

                # add the output file info to the dictionary
                output_jobrep[name] = nentries

        # now make sure that the known output files are in the job report dictionary
        for lfn in lfns_jobdef:
            if lfn not in output_jobrep and lfn not in job.allownooutput:
                log.warning('output file %s from job definition is not present in job report and is not listed in allowNoOutput - job will fail' % lfn)
                job.piloterrorcodes, job.piloterrordiags = errors.add_error_code(errors.MISSINGOUTPUTFILE)
                failed = True
                break
            if lfn not in output_jobrep and lfn in job.allownooutput:
                log.warning('output file %s from job definition is not present in job report but is listed in allowNoOutput - remove from stage-out' % lfn)
                remove_from_stageout(lfn, job)
            else:
                nentries = output_jobrep[lfn]
                if nentries is not None and nentries == 0 and lfn not in job.allownooutput:
                    log.warning('output file %s is listed in job report, has zero events and is not listed in allowNoOutput - job will fail' % lfn)
                    job.piloterrorcodes, job.piloterrordiags = errors.add_error_code(errors.EMPTYOUTPUTFILE)
                    failed = True
                    break
                if nentries == 0 and lfn in job.allownooutput:
                    log.warning('output file %s is listed in job report, nentries=0 and is listed in allowNoOutput - remove from stage-out' % lfn)
                    remove_from_stageout(lfn, job)
                elif not nentries and lfn not in job.allownooutput:
                    log.warning('output file %s is listed in job report, nentries is not set and is not listed in allowNoOutput - ignore' % lfn)
                elif not nentries and lfn in job.allownooutput:
                    log.warning('output file %s is listed in job report, nentries is None but is listed in allowNoOutput - remove from stage-out' % lfn)
                    remove_from_stageout(lfn, job)
                elif nentries:
                    log.info('output file %s has %d events' % (lfn, nentries))
                else:  # should not reach this step
                    log.warning('case not handled for output file %s with %s events (ignore)' % (lfn, str(nentries)))

    status = True if not failed else False

    if status:
        log.info('output file verification succeeded')
    else:
        log.warning('output file verification failed')

    return status


def remove_from_stageout(lfn, job):
    """
    From the given lfn from the stage-out list.

    :param lfn: local file name (string).
    :param job: job object
    :return: [updated job object]
    """

    log = get_logger(job.jobid)
    outdata = []
    for fspec in job.outdata:
        if fspec.lfn == lfn:
            log.info('removing %s from stage-out list' % lfn)
        else:
            outdata.append(fspec)
    job.outdata = outdata


def remove_no_output_files(job):
    """
    Remove files from output file list if they are listed in allowNoOutput and do not exist.

    :param job: job object.
    :return:
    """

    log = get_logger(job.jobid)

    # first identify the files to keep
    _outfiles = []
    for fspec in job.outdata:
        filename = fspec.lfn
        path = os.path.join(job.workdir, filename)

        if filename in job.allownooutput:
            if os.path.exists(path):
                log.info("file %s is listed in allowNoOutput but exists (will not be removed from list of files to be staged-out)" % filename)
                _outfiles.append(filename)
            else:
                log.info("file %s is listed in allowNoOutput and does not exist (will be removed from list of files to be staged-out)" % filename)
        else:
            if os.path.exists(path):
                log.info("file %s is not listed in allowNoOutput (will be staged-out)" % filename)
            else:
                log.warning("file %s is not listed in allowNoOutput and does not exist (job will fail)" % filename)
            _outfiles.append(filename)

    # now remove the unwanted fspecs
    if len(_outfiles) != len(job.outdata):
        outdata = []
        for fspec in job.outdata:
            if fspec.lfn in _outfiles:
                outdata.append(fspec)
        job.outdata = outdata


def get_outfiles_records(subfiles):
    """
    Extract file info from job report JSON subfiles entry.

    :param subfiles: list of subfiles.
    :return: file info dictionary with format { 'guid': .., 'size': .., 'nentries': .. (optional)}
    """

    res = {}
    for f in subfiles:
        res[f['name']] = {'guid': f['file_guid'],
                          'size': f['file_size']}
        nentries = f.get('nentries', 'UNDEFINED')
        if type(nentries) == int:
            res[f['name']]['nentries'] = nentries
        else:
            logger.warning("nentries is undefined in job report")

    return res


class DictQuery(dict):
    def get(self, path, dst_dict, dst_key):
        keys = path.split("/")
        if len(keys) == 0:
            return
        last_key = keys.pop()
        v = self
        for key in keys:
            if key in v and isinstance(v[key], dict):
                v = v[key]
            else:
                return
        if last_key in v:
            dst_dict[dst_key] = v[last_key]


def parse_jobreport_data(job_report):
    """
    Parse a job report and extract relevant fields.

    :param job_report:
    :return:
    """
    work_attributes = {}
    if job_report is None or not any(job_report):
        return work_attributes

    # these are default values for job metrics
    core_count = ""
    work_attributes["nEvents"] = 0
    work_attributes["dbTime"] = ""
    work_attributes["dbData"] = ""
    work_attributes["inputfiles"] = []
    work_attributes["outputfiles"] = []

    if "ATHENA_PROC_NUMBER" in os.environ:
        logger.debug("ATHENA_PROC_NUMBER: {0}".format(os.environ["ATHENA_PROC_NUMBER"]))
        work_attributes['core_count'] = int(os.environ["ATHENA_PROC_NUMBER"])
        core_count = int(os.environ["ATHENA_PROC_NUMBER"])

    dq = DictQuery(job_report)
    dq.get("resource/transform/processedEvents", work_attributes, "nEvents")
    dq.get("resource/transform/cpuTimeTotal", work_attributes, "cpuConsumptionTime")
    dq.get("resource/machine/node", work_attributes, "node")
    dq.get("resource/machine/model_name", work_attributes, "cpuConsumptionUnit")
    dq.get("resource/dbTimeTotal", work_attributes, "dbTime")
    dq.get("resource/dbDataTotal", work_attributes, "dbData")
    dq.get("exitCode", work_attributes, "transExitCode")
    dq.get("exitMsg", work_attributes, "exeErrorDiag")
    dq.get("files/input", work_attributes, "inputfiles")
    dq.get("files/output", work_attributes, "outputfiles")

    outputfiles_dict = {}
    for of in work_attributes['outputfiles']:
        outputfiles_dict.update(get_outfiles_records(of['subFiles']))
    work_attributes['outputfiles'] = outputfiles_dict

    if work_attributes['inputfiles']:
        work_attributes['nInputFiles'] = reduce(lambda a, b: a + b, map(lambda inpfiles: len(inpfiles['subFiles']),
                                                                        work_attributes['inputfiles']))

    if 'resource' in job_report and 'executor' in job_report['resource']:
        j = job_report['resource']['executor']
        exc_report = []
        fin_report = defaultdict(int)
        for v in filter(lambda d: 'memory' in d and ('Max' or 'Avg' in d['memory']), j.itervalues()):
            if 'Avg' in v['memory']:
                exc_report.extend(v['memory']['Avg'].items())
            if 'Max' in v['memory']:
                exc_report.extend(v['memory']['Max'].items())
        for x in exc_report:
            fin_report[x[0]] += x[1]
        work_attributes.update(fin_report)

    workdir_size = get_workdir_size()
    work_attributes['jobMetrics'] = 'coreCount=%s nEvents=%s dbTime=%s dbData=%s workDirSize=%s' % \
                                    (core_count,
                                        work_attributes["nEvents"],
                                        work_attributes["dbTime"],
                                        work_attributes["dbData"],
                                        workdir_size)
    del(work_attributes["dbData"])
    del(work_attributes["dbTime"])

    return work_attributes


def get_workdir_size():
    """
    Tmp function - move later to file_handling

    :return:
    """
    c, o, e = execute('du -s', shell=True)
    if o is not None:
        return o.split()[0]
    return None


def get_executor_dictionary(jobreport_dictionary):
    """
    Extract the 'executor' dictionary from with a job report.

    :param jobreport_dictionary:
    :return: executor_dictionary
    """

    executor_dictionary = {}
    if jobreport_dictionary != {}:

        if 'resource' in jobreport_dictionary:
            resource_dictionary = jobreport_dictionary['resource']
            if 'executor' in resource_dictionary:
                executor_dictionary = resource_dictionary['executor']
            else:
                logger.warning("no such key: executor")
        else:
            logger.warning("no such key: resource")

    return executor_dictionary


def get_number_of_events_deprecated(jobreport_dictionary):  # TODO: remove this function
    """
    Extract the number of events from the job report.

    :param jobreport_dictionary:
    :return:
    """

    nevents = {}  # FORMAT: { format : total_events, .. }
    nmax = 0

    executor_dictionary = get_executor_dictionary(jobreport_dictionary)
    if executor_dictionary != {}:
        for format in executor_dictionary.keys():  # "RAWtoESD", ..
            if 'nevents' in executor_dictionary[format]:
                if format in nevents:
                    nevents[format] += executor_dictionary[format]['nevents']
                else:
                    nevents[format] = executor_dictionary[format]['nevents']
            else:
                logger.warning("format %s has no such key: nevents" % (format))

    # Now find the largest number of events among the different formats
    if nevents != {}:
        try:
            nmax = max(nevents.values())
        except Exception as e:
            logger.warning("exception caught: %s" % (e))
            nmax = 0
    else:
        logger.warning("did not find the number of events in the job report")
        nmax = 0

    return nmax


def get_db_info(jobreport_dictionary):
    """
    Extract and add up the DB info from the job report.
    This information is reported with the jobMetrics.
    Note: this function adds up the different dbData and dbTime's in the different executor steps. In modern job
    reports this might have been done already by the transform and stored in dbDataTotal and dbTimeTotal.

    :param jobreport_dictionary: job report dictionary.
    :return: db_time (int), db_data (long)
    """

    db_time = 0
    db_data = 0L

    executor_dictionary = get_executor_dictionary(jobreport_dictionary)
    if executor_dictionary != {}:
        for format in executor_dictionary.keys():  # "RAWtoESD", ..
            if 'dbData' in executor_dictionary[format]:
                try:
                    db_data += executor_dictionary[format]['dbData']
                except Exception:
                    pass
            else:
                logger.warning("format %s has no such key: dbData" % (format))
            if 'dbTime' in executor_dictionary[format]:
                try:
                    db_time += executor_dictionary[format]['dbTime']
                except Exception:
                    pass
            else:
                logger.warning("format %s has no such key: dbTime" % (format))

    return db_time, db_data


def get_db_info_str(db_time, db_data):
    """
    Convert db_time, db_data to strings.
    E.g. dbData="105077960", dbTime="251.42".

    :param db_time: time (s)
    :param db_data: long integer
    :return: db_time_s, db_data_s (strings)
    """

    if db_data != 0L:
        db_data_s = "%s" % (db_data)
    else:
        db_data_s = ""
    if db_time != 0:
        db_time_s = "%.2f" % (db_time)
    else:
        db_time_s = ""

    return db_time_s, db_data_s


def get_cpu_times(jobreport_dictionary):
    """
    Extract and add up the total CPU times from the job report.
    E.g. ('s', 5790L, 1.0).

    Note: this function is used with Event Service jobs

    :param jobreport_dictionary:
    :return: cpu_conversion_unit (unit), total_cpu_time, conversion_factor (output consistent with set_time_consumed())
    """

    total_cpu_time = 0L

    executor_dictionary = get_executor_dictionary(jobreport_dictionary)
    if executor_dictionary != {}:
        for format in executor_dictionary.keys():  # "RAWtoESD", ..
            if 'cpuTime' in executor_dictionary[format]:
                try:
                    total_cpu_time += executor_dictionary[format]['cpuTime']
                except Exception:
                    pass
            else:
                logger.warning("format %s has no such key: cpuTime" % (format))

    conversion_factor = 1.0
    cpu_conversion_unit = "s"

    return cpu_conversion_unit, total_cpu_time, conversion_factor


def get_exit_info(jobreport_dictionary):
    """
    Return the exit code (exitCode) and exit message (exitMsg).
    E.g. (0, 'OK').

    :param jobreport_dictionary:
    :return: exit_code, exit_message
    """

    return jobreport_dictionary['exitCode'], jobreport_dictionary['exitMsg']


def cleanup_payload(workdir, outputfiles=[]):
    """
    Cleanup of payload (specifically AthenaMP) sub directories prior to log file creation.

    :param workdir: working directory (string)
    :param outputfiles: list of output files
    :return:
    """

    for ampdir in glob('%s/athenaMP-workers-*' % (workdir)):
        for (p, d, f) in os.walk(ampdir):
            for filename in f:
                if 'core' in filename or 'pool.root' in filename or 'tmp.' in filename:
                    path = os.path.join(p, filename)
                    path = os.path.abspath(path)
                    remove(path)
                for outfile in outputfiles:
                    if outfile in filename:
                        path = os.path.join(p, filename)
                        path = os.path.abspath(path)
                        remove(path)


def get_redundant_path():
    """
    Return the path to the file containing the redundant files and directories to be removed prior to log file creation.

    :return: file path (string).
    """

    filename = config.Pilot.redundant

    # correct /cvmfs if necessary
    if filename.startswith('/cvmfs') and os.environ.get('ATLAS_SW_BASE', False):
        filename = filename.replace('/cvmfs', os.environ.get('ATLAS_SW_BASE'))

    return filename


def get_redundants():
    """
    Get list of redundant files and directories (to be removed).
    The function will return the content of an external file. It that can't be read, then a list defined in this
    function will be returned instead. Any updates to the external file must be propagated to this function.

    :return: files and directories list
    """

    # try to read the list from the external file
    filename = get_redundant_path()
    if os.path.exists(filename):
        dir_list = read_list(filename)
        if dir_list:
            return dir_list

    logger.debug('list of redundant files could not be read from external file: %s (will use internal list)' % filename)

    # else return the following
    dir_list = ["AtlasProduction*",
                "AtlasPoint1",
                "AtlasTier0",
                "buildJob*",
                "CDRelease*",
                "csc*.log",
                "DBRelease*",
                "EvgenJobOptions",
                "external",
                "fort.*",
                "geant4",
                "geomDB",
                "geomDB_sqlite",
                "home",
                "o..pacman..o",
                "pacman-*",
                "python",
                "runAthena*",
                "share",
                "sources.*",
                "sqlite*",
                "sw",
                "tcf_*",
                "triggerDB",
                "trusted.caches",
                "workdir",
                "*.data*",
                "*.events",
                "*.py",
                "*.pyc",
                "*.root*",
                "JEM",
                "tmp*",
                "*.tmp",
                "*.TMP",
                "MC11JobOptions",
                "scratch",
                "*.writing",
                "pwg*",
                "pwhg*",
                "*PROC*",
                "madevent",
                "*proxy",
                "ckpt*",
                "*runcontainer*",
                "*job.log.tgz",
                "runGen-*",
                "runAthena-*",
                "pandawnutil/*",
                "src/*",
                "singularity_cachedir",
                "_joproxy15",
                "HAHM_*",
                "Process",
                "merged_lhef._0.events-new"]

    return dir_list


def remove_archives(workdir):
    """
    Explicitly remove any soft linked archives (.a files) since they will be dereferenced by the tar command
    (--dereference option).

    :param workdir: working directory (string)
    :return:
    """

    matches = []
    for root, dirnames, filenames in os.walk(workdir):
        for filename in fnmatch.filter(filenames, '*.a'):
            matches.append(os.path.join(root, filename))
    for root, dirnames, filenames in os.walk(os.path.dirname(workdir)):
        for filename in fnmatch.filter(filenames, 'EventService_premerge_*.tar'):
            matches.append(os.path.join(root, filename))
    if matches != []:
        for f in matches:
            remove(f)


def cleanup_broken_links(workdir):
    """
    Run a second pass to clean up any broken links prior to log file creation.

    :param workdir: working directory (string)
    :return:
    """

    broken = []
    for root, dirs, files in os.walk(workdir):
        for filename in files:
            path = os.path.join(root, filename)
            if os.path.islink(path):
                target_path = os.readlink(path)
                # Resolve relative symlinks
                if not os.path.isabs(target_path):
                    target_path = os.path.join(os.path.dirname(path), target_path)
                if not os.path.exists(target_path):
                    broken.append(path)
            else:
                # If it's not a symlink we're not interested.
                continue

    if broken:
        for p in broken:
            remove(p)


def remove_redundant_files(workdir, outputfiles=[]):
    """
    Remove redundant files and directories prior to creating the log file.

    :param workdir: working directory (string).
    :param outputfiles: list of output files.
    :return:
    """

    logger.debug("removing redundant files prior to log creation")

    workdir = os.path.abspath(workdir)

    # get list of redundant files and directories (to be removed)
    dir_list = get_redundants()

    # remove core and pool.root files from AthenaMP sub directories
    try:
        cleanup_payload(workdir, outputfiles)
    except Exception as e:
        logger.warning("failed to execute cleanup_payload(): %s" % e)

    # explicitly remove any soft linked archives (.a files) since they will be dereferenced by the tar command
    # (--dereference option)
    remove_archives(workdir)

    # note: these should be partial file/dir names, not containing any wildcards
    exceptions_list = ["runargs", "runwrapper", "jobReport", "log."]

    to_delete = []
    for _dir in dir_list:
        files = glob(os.path.join(workdir, _dir))
        exclude = []

        if files:
            for exc in exceptions_list:
                for f in files:
                    if exc in f:
                        exclude.append(os.path.abspath(f))

            _files = []
            for f in files:
                if f not in exclude:
                    _files.append(os.path.abspath(f))
            to_delete += _files

    exclude_files = []
    for of in outputfiles:
        exclude_files.append(os.path.join(workdir, of))
    for f in to_delete:
        if f not in exclude_files:
            if os.path.isfile(f):
                remove(f)
            else:
                remove_dir_tree(f)

    # run a second pass to clean up any broken links
    cleanup_broken_links(workdir)

    # remove any present user workDir
    path = os.path.join(workdir, 'workDir')
    if os.path.exists(path):
        remove_dir_tree(path)


def get_utility_commands_list(order=None):
    """
    Return a list of utility commands to be executed in parallel with the payload.
    This could e.g. be memory and network monitor commands. A separate function can be used to determine the
    corresponding command setups using the utility command name.
    If the optional order parameter is set, the function should return the list of corresponding commands.
    E.g. if order=UTILITY_BEFORE_PAYLOAD, the function should return all commands that are to be executed before the
    payload. If order=UTILITY_WITH_PAYLOAD, the corresponding commands will be prepended to the payload execution
    string. If order=UTILITY_AFTER_PAYLOAD_STARTED, the commands that should be executed after the payload has been started
    should be returned. If order=UTILITY_WITH_STAGEIN, the commands that should be executed parallel with stage-in will
    be returned.

    :param order: optional sorting order (see pilot.util.constants)
    :return: list of utilities to be executed in parallel with the payload.
    """

    if order:
        if order == UTILITY_BEFORE_PAYLOAD:
            return ['Prefetcher']
        elif order == UTILITY_WITH_PAYLOAD:
            return ['NetworkMonitor']
        elif order == UTILITY_AFTER_PAYLOAD_STARTED:
            return ['MemoryMonitor']
        elif order == UTILITY_WITH_STAGEIN:
            return ['Benchmark']
    return []


def get_utility_command_setup(name, job, setup=None):
    """
    Return the proper setup for the given utility command.
    If a payload setup is specified, then the utility command string should be prepended to it.

    :param name: name of utility (string).
    :param job: job object.
    :param setup: optional payload setup string.
    :return: utility command setup (string).
    """

    if name == 'MemoryMonitor':
        setup, pid = get_memory_monitor_setup(job.pid, job.pgrp, job.jobid, job.workdir, job.command, use_container=job.usecontainer,
                                              transformation=job.transformation, outdata=job.outdata)
        _pattern = r"([\S]+)\ ."
        pattern = re.compile(_pattern)
        _name = re.findall(pattern, setup.split(';')[-1])
        if _name:
            job.memorymonitor = _name[0]
        else:
            logger.warning('trf name could not be identified in setup string')

        # update the pgrp if the pid changed
        if job.pid != pid and pid != --1:
            logger.debug('updating pgrp=%d for pid=%d' % (job.pgrp, pid))
            try:
                job.pgrp = os.getpgid(pid)
            except Exception as e:
                logger.warning('os.getpgid(%d) failed with: %s' % (pid, e))
        return setup
    elif name == 'NetworkMonitor' and setup:
        return get_network_monitor_setup(setup, job)
    elif name == 'Prefetcher':
        return get_prefetcher_setup(job)
    elif name == 'Benchmark':
        return get_benchmark_setup(job)
    else:
        return ""


def get_utility_command_execution_order(name):
    """
    Should the given utility command be executed before or after the payload?

    :param name: utility name (string).
    :return: execution order constant.
    """

    # example implementation
    if name == 'NetworkMonitor':
        return UTILITY_WITH_PAYLOAD
    elif name == 'MemoryMonitor':
        return UTILITY_AFTER_PAYLOAD_STARTED
    else:
        logger.warning('unknown utility name: %s' % name)
        return UTILITY_AFTER_PAYLOAD_STARTED


def post_utility_command_action(name, job):
    """
    Perform post action for given utility command.

    :param name: name of utility command (string).
    :param job: job object.
    :return:
    """

    if name == 'NetworkMonitor':
        pass
    elif name == 'MemoryMonitor':
        post_memory_monitor_action(job)


def get_utility_command_kill_signal(name):
    """
    Return the proper kill signal used to stop the utility command.

    :param name:
    :return: kill signal
    """

    if name == 'MemoryMonitor':
        sig = SIGUSR1
    else:
        # note that the NetworkMonitor does not require killing (to be confirmed)
        sig = SIGTERM

    return sig


def get_utility_command_output_filename(name, selector=None):
    """
    Return the filename to the output of the utility command.

    :param name: utility name (string).
    :param selector: optional special conditions flag (boolean).
    :return: filename (string).
    """

    if name == 'MemoryMonitor':
        filename = get_memory_monitor_summary_filename(selector=selector)
    else:
        filename = ""

    return filename


def verify_lfn_length(outdata):
    """
    Make sure that the LFNs are all within the allowed length.

    :param outdata: FileSpec object.
    :return: error code (int), diagnostics (string).
    """

    ec = 0
    diagnostics = ""
    max_length = 255

    # loop over all output files
    for fspec in outdata:
        if len(fspec.lfn) > max_length:
            diagnostics = "LFN too long (length: %d, must be less than %d characters): %s" % \
                          (len(fspec.lfn), max_length, fspec.lfn)
            ec = errors.LFNTOOLONG
            break

    return ec, diagnostics


def verify_ncores(corecount):
    """
    Verify that nCores settings are correct

    :param corecount: number of cores (int).
    :return:
    """

    try:
        del os.environ['ATHENA_PROC_NUMBER_JOB']
        logger.debug("unset existing ATHENA_PROC_NUMBER_JOB")
    except Exception:
        pass

    try:
        athena_proc_number = int(os.environ.get('ATHENA_PROC_NUMBER', None))
    except Exception:
        athena_proc_number = None

    # Note: if ATHENA_PROC_NUMBER is set (by the wrapper), then do not overwrite it
    # Otherwise, set it to the value of job.coreCount
    # (actually set ATHENA_PROC_NUMBER_JOB and use it if it exists, otherwise use ATHENA_PROC_NUMBER directly;
    # ATHENA_PROC_NUMBER_JOB will always be the value from the job definition)
    if athena_proc_number:
        logger.info("encountered a set ATHENA_PROC_NUMBER (%d), will not overwrite it" % athena_proc_number)
    else:
        os.environ['ATHENA_PROC_NUMBER_JOB'] = "%s" % corecount
        logger.info("set ATHENA_PROC_NUMBER_JOB to %s (ATHENA_PROC_NUMBER will not be overwritten)" % corecount)


def verify_job(job):
    """
    Verify job parameters for specific errors.
    Note:
      in case of problem, the function should set the corresponding pilot error code using
      job.piloterrorcodes, job.piloterrordiags = errors.add_error_code(error.get_error_code())

    :param job: job object
    :return: Boolean.
    """

    status = False
    log = get_logger(job.jobid)

    # are LFNs of correct lengths?
    ec, diagnostics = verify_lfn_length(job.outdata)
    if ec != 0:
        log.fatal(diagnostics)
        job.piloterrordiag = diagnostics
        job.piloterrorcodes, job.piloterrordiags = errors.add_error_code(ec)
    else:
        status = True

    # check the ATHENA_PROC_NUMBER settings
    verify_ncores(job.corecount)

    return status



#!/usr/bin/env python
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Mario Lassnig, mario.lassnig@cern.ch, 2017
# - Paul Nilsson, paul.nilsson@cern.ch, 2017-2019
# - Tobias Wegner, tobias.wegner@cern.ch, 2017-2018
# - Alexey Anisenkov, anisyonk@cern.ch, 2018

# refactored by Alexey Anisenkov

import os
import hashlib
import logging
import time

from pilot.info import infosys
from pilot.info.storageactivitymaps import get_ddm_activity
from pilot.common.exception import PilotException, ErrorCodes, SizeTooLarge, NoLocalSpace, ReplicasNotFound
from pilot.util.filehandling import calculate_checksum
from pilot.util.math import convert_mb_to_b
from pilot.util.parameters import get_maximum_input_sizes
from pilot.util.workernode import get_local_disk_space
from pilot.util.timer import TimeoutException
from pilot.util.tracereport import TraceReport

errors = ErrorCodes()


class StagingClient(object):
    """
        Base Staging Client
    """

    mode = ""  # stage-in/out, set by the inheritor of the class
    copytool_modules = {'rucio': {'module_name': 'rucio'},
                        'gfal': {'module_name': 'gfal'},
                        'gfalcopy': {'module_name': 'gfal'},
                        'xrdcp': {'module_name': 'xrdcp'},
                        'mv': {'module_name': 'mv'},
                        'lsm': {'module_name': 'lsm'}
                        }

    # list of allowed schemas to be used for direct acccess mode from REMOTE replicas
    direct_remoteinput_allowed_schemas = ['root']
    # list of schemas to be used for direct acccess mode from LOCAL replicas
    direct_localinput_allowed_schemas = ['root', 'dcache', 'dcap', 'file', 'https']
    # list of allowed schemas to be used for transfers from REMOTE sites
    remoteinput_allowed_schemas = ['root', 'gsiftp', 'dcap', 'davs', 'srm', 'storm']

    def __init__(self, infosys_instance=None, acopytools=None, logger=None, default_copytools='rucio', trace_report=None):
        """
            If `acopytools` is not specified then it will be automatically resolved via infosys. In this case `infosys` requires initialization.
            :param acopytools: dict of copytool names per activity to be used for transfers. Accepts also list of names or string value without activity passed.
            :param logger: logging.Logger object to use for logging (None means no logging)
            :param default_copytools: copytool name(s) to be used in case of unknown activity passed. Accepts either list of names or single string value.
        """

        super(StagingClient, self).__init__()

        if not logger:
            logger = logging.getLogger('%s.%s' % (__name__, 'null'))
            logger.disabled = True

        self.logger = logger
        self.infosys = infosys_instance or infosys

        if isinstance(acopytools, basestring):
            acopytools = {'default': [acopytools]} if acopytools else {}
        if isinstance(acopytools, (list, tuple)):
            acopytools = {'default': acopytools} if acopytools else {}

        self.acopytools = acopytools or {}

        if self.infosys.queuedata:
            if not self.acopytools:  ## resolve from queuedata.acopytools using infosys
                self.acopytools = (self.infosys.queuedata.acopytools or {}).copy()
            if not self.acopytools:  ## resolve from queuedata.copytools using infosys
                self.acopytools = dict(default=(self.infosys.queuedata.copytools or {}).keys())

        if not self.acopytools.get('default'):
            if isinstance(default_copytools, basestring):
                default_copytools = [default_copytools] if default_copytools else []
            self.acopytools['default'] = default_copytools

        if not self.acopytools:
            msg = 'failed to initilize StagingClient: no acopytools options found, acopytools=%s' % self.acopytools
            logger.error(msg)
            self.trace_report.update(clientState='BAD_COPYTOOL', stateReason=msg)
            self.trace_report.send()
            raise PilotException("failed to resolve acopytools settings")
        logger.info('configured copytools per activity: acopytools=%s' % self.acopytools)

        # get an initialized trace report (has to be updated for get/put if not defined before)
        self.trace_report = trace_report if trace_report else TraceReport(pq=os.environ.get('PILOT_SITENAME', ''))

    @classmethod
    def get_preferred_replica(self, replicas, allowed_schemas):
        """
            Get preferred replica from the `replicas` list suitable for `allowed_schemas`
            :return: replica or None if not found
        """

        for replica in replicas:
            for schema in allowed_schemas:
                if replica and (not schema or replica.startswith('%s://' % schema)):
                    return replica

    def resolve_replicas(self, files):  # noqa: C901
        """
            Populates filespec.replicas for each entry from `files` list
            :param files: list of `FileSpec` objects
            fdat.replicas = [(ddmendpoint, replica, ddm_se, ddm_path)]
            :return: `files`
        """

        logger = self.logger  ## the function could be static if logger will be moved outside

        xfiles = []
        #ddmconf = self.infosys.resolve_storage_data()

        for fdat in files:
            #ddmdat = ddmconf.get(fdat.ddmendpoint)
            #if not ddmdat:
            #    raise Exception("Failed to resolve input ddmendpoint by name=%s (from PanDA), please check configuration. fdat=%s" % (fdat.ddmendpoint, fdat))

            ## skip fdat if need for further workflow (e.g. to properly handle OS ddms)

            #fdat.accessmode = 'copy'        ### quick hack to avoid changing logic below for DIRECT access handling  ## REVIEW AND FIX ME LATER
            #fdat.allowremoteinputs = False  ### quick hack to avoid changing logic below for DIRECT access handling  ## REVIEW AND FIX ME LATER

            if not fdat.inputddms and self.infosys.queuedata:
                fdat.inputddms = self.infosys.queuedata.astorages.get('pr', {})  ## FIX ME LATER: change to proper activity=read_lan
            if not fdat.inputddms and fdat.ddmendpoint:
                fdat.inputddms = [fdat.ddmendpoint]
            xfiles.append(fdat)

        if not xfiles:  # no files for replica look-up
            return files

        # load replicas from Rucio
        from rucio.client import Client
        c = Client()

        ## for the time being until Rucio bug with geo-ip sorting is resolved
        ## do apply either simple query list_replicas() without geoip sort to resolve LAN replicas in case of directaccesstype=[None, LAN]
        # otherwise in case of directaccesstype=WAN mode do query geo sorted list_replicas() with location data passed

        bquery = {'schemes': ['srm', 'root', 'davs', 'gsiftp', 'https', 'storm'],
                  'dids': [dict(scope=e.scope, name=e.lfn) for e in xfiles]}

        #allow_remoteinput = True in set(e.allowremoteinputs for e in xfiles)  ## implement direct access later

        try:
            query = bquery.copy()
            location = self.detect_client_location()
            if not location:
                raise PilotException("Failed to get client location for Rucio", code=errors.RUCIOLOCATIONFAILED)

            query.update(sort='geoip', client_location=location)
            logger.info('calling rucio.list_replicas() with query=%s' % query)
            replicas = c.list_replicas(**query)
        except Exception as e:
            raise PilotException("Failed to get replicas from Rucio: %s" % e, code=errors.RUCIOLISTREPLICASFAILED)

        replicas = list(replicas)
        logger.debug("replicas received from Rucio: %s" % replicas)

        files_lfn = dict(((e.scope, e.lfn), e) for e in xfiles)
        logger.debug("files_lfn=%s" % files_lfn)

        for r in replicas:
            k = r['scope'], r['name']
            fdat = files_lfn.get(k)
            if not fdat:  # not requested replica returned?
                continue

            fdat.replicas = []  # reset replicas list
            has_direct_localinput_replicas = False

            # manually sort replicas by priority value since Rucio has a bug in ordering
            # .. can be removed once Rucio server-side fix will be delivered
            ordered_replicas = {}
            for pfn, xdat in sorted(r.get('pfns', {}).iteritems(), key=lambda x: x[1]['priority']):
                ordered_replicas.setdefault(xdat.get('rse'), []).append(pfn)

            # local replicas
            for ddm in fdat.inputddms:  ## iterate over local ddms and check if replica is exist here
                #pfns = r.get('rses', {}).get(ddm)  ## use me once Rucio bug is resolved
                pfns = ordered_replicas.get(ddm)    ## quick workaround of Rucio bug: use manually sorted pfns

                if not pfns:  # no replica found for given local ddm
                    continue

                fdat.replicas.append((ddm, pfns))

                if not has_direct_localinput_replicas:
                    has_direct_localinput_replicas = bool(self.get_preferred_replica(pfns, self.direct_localinput_allowed_schemas))

            if (not fdat.replicas or (fdat.accessmode == 'direct' and not has_direct_localinput_replicas)) and fdat.allowremoteinputs:
                if fdat.accessmode == 'direct':
                    allowed_schemas = self.direct_remoteinput_allowed_schemas
                else:
                    allowed_schemas = self.remoteinput_allowed_schemas

                if not fdat.replicas:
                    logger.info("No local replicas found for lfn=%s but allowremoteinputs is set => looking for remote inputs" % fdat.lfn)
                else:
                    logger.info("Direct access=True but no appr. local replicas found: allowremoteinputs=True =>checking remote input suitable for direct read")
                logger.info("Consider first/closest replica, accessmode=%s, allowed_schemas=%s" % (fdat.accessmode, allowed_schemas))
                #logger.debug('rses=%s' % r['rses'])
                for ddm, pfns in r['rses'].iteritems():
                    replica = self.get_preferred_replica(pfns, allowed_schemas)
                    if not replica:
                        continue

                    # remoteinput supported replica found
                    fdat.replicas.append((ddm, pfns))
                    # break # ignore other remote replicas/sites

            # verify filesize and checksum values
            self.trace_report.update(validateStart=time.time())
            status = True
            if fdat.filesize != r['bytes']:
                logger.warning("Filesize of input file=%s mismatched with value from Rucio replica: filesize=%s, replica.filesize=%s, fdat=%s"
                               % (fdat.lfn, fdat.filesize, r['bytes'], fdat))
                status = False

            if not fdat.filesize:
                fdat.filesize = r['bytes']
                logger.warning("Filesize value for input file=%s is not defined, assigning info from Rucio replica: filesize=%s" % (fdat.lfn, r['bytes']))

            for ctype in ['adler32', 'md5']:
                if fdat.checksum.get(ctype) != r[ctype] and r[ctype]:
                    logger.warning("Checksum value of input file=%s mismatched with info got from Rucio replica: checksum=%s, replica.checksum=%s, fdat=%s"
                                   % (fdat.lfn, fdat.checksum, r[ctype], fdat))
                    status = False

                if not fdat.checksum.get(ctype) and r[ctype]:
                    fdat.checksum[ctype] = r[ctype]

            if not status:
                logger.info("filesize and checksum verification done")
                self.trace_report.update(clientState="DONE")

        logger.info('Number of resolved replicas:\n' +
                    '\n'.join(["lfn=%s: replicas=%s, allowremoteinputs=%s, is_directaccess=%s"
                               % (f.lfn, len(f.replicas), f.allowremoteinputs, f.is_directaccess(ensure_replica=False)) for f in files]))

        return files

    @classmethod
    def detect_client_location(self):
        """
        Open a UDP socket to a machine on the internet, to get the local IPv4 and IPv6
        addresses of the requesting client.
        Try to determine the sitename automatically from common environment variables,
        in this order: SITE_NAME, ATLAS_SITE_NAME, OSG_SITE_NAME. If none of these exist
        use the fixed string 'ROAMING'.
        """

        ip = '0.0.0.0'
        try:
            import socket
            s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            s.connect(("8.8.8.8", 80))
            ip = s.getsockname()[0]
        except Exception:
            pass

        ip6 = '::'
        try:
            s = socket.socket(socket.AF_INET6, socket.SOCK_DGRAM)
            s.connect(("2001:4860:4860:0:0:0:0:8888", 80))
            ip6 = s.getsockname()[0]
        except Exception:
            pass

        site = os.environ.get('PILOT_RUCIO_SITENAME', 'unknown')
#        site = os.environ.get('SITE_NAME',
#                              os.environ.get('ATLAS_SITE_NAME',
#                                             os.environ.get('OSG_SITE_NAME',
#                                                            'ROAMING')))

        return {'ip': ip,
                'ip6': ip6,
                'fqdn': socket.getfqdn(),
                'site': site}

    def transfer_files(self, copytool, files, **kwargs):
        """
            Apply transfer of given `files` using passed `copytool` module
            Should be implemented by custom Staging Client
            :param copytool: copytool module
            :param files: list of `FileSpec` objects
            :param kwargs: extra kwargs to be passed to copytool transfer handler
            :raise: PilotException in case of controlled error
        """

        raise NotImplemented()

    def transfer(self, files, activity='default', **kwargs):  # noqa: C901
        """
            Automatically stage passed files using copy tools related to given `activity`
            :param files: list of `FileSpec` objects
            :param activity: list of activity names used to determine appropriate copytool (prioritized list)
            :param kwargs: extra kwargs to be passed to copytool transfer handler
            :raise: PilotException in case of controlled error
            :return: output of copytool transfers (to be clarified)
        """

        self.trace_report.update(relativeStart=time.time(), transferStart=time.time())

        if isinstance(activity, basestring):
            activity = [activity]
        if 'default' not in activity:
            activity.append('default')

        # stage-in/out?
        #mode = kwargs.get('mode', '')

        copytools = None
        for aname in activity:
            copytools = self.acopytools.get(aname)
            if copytools:
                break

        if not copytools:
            raise PilotException('failed to resolve copytool by preferred activities=%s, acopytools=%s' %
                                 (activity, self.acopytools))

        result, caught_errors = None, []

        for name in copytools:

            try:
                if name not in self.copytool_modules:
                    raise PilotException('passed unknown copytool with name=%s .. skipped' % name,
                                         code=ErrorCodes.UNKNOWNCOPYTOOL)

                module = self.copytool_modules[name]['module_name']
                self.logger.info('trying to use copytool=%s for activity=%s' % (name, activity))
                copytool = __import__('pilot.copytool.%s' % module, globals(), locals(), [module], -1)
                self.trace_report.update(protocol=name)

            except PilotException as e:
                caught_errors.append(e)
                self.logger.debug('error: %s' % e)
                continue
            except Exception as e:
                self.logger.warning('failed to import copytool module=%s, error=%s' % (module, e))
                continue
            try:
                result = self.transfer_files(copytool, files, activity, **kwargs)
            except PilotException as e:
                msg = 'failed to execute transfer_files(): PilotException caught: %s' % e
                self.logger.warning(msg)
                caught_errors.append(e)
            except TimeoutException as e:
                msg = 'function timed out: %s' % e
                self.logger.warning(msg)
                caught_errors.append(e)
            except Exception as e:
                self.logger.warning('failed to transfer files using copytool=%s .. skipped; error=%s' % (copytool, e))
                import traceback
                self.logger.error(traceback.format_exc())
                caught_errors.append(e)
            else:
                self.logger.debug('transfer_files() completed with result=%s' % str(result))

            if caught_errors and isinstance(caught_errors[-1], PilotException) and \
                    caught_errors[-1].get_error_code() == ErrorCodes.MISSINGOUTPUTFILE:
                raise caught_errors[-1]

            if result:
                break

        if not result:
            if caught_errors and "Cannot authenticate" in str(caught_errors):
                code = ErrorCodes.STAGEINAUTHENTICATIONFAILURE
            elif caught_errors and "bad queue configuration" in str(caught_errors):
                code = ErrorCodes.BADQUEUECONFIGURATION
            elif caught_errors and isinstance(caught_errors[-1], PilotException):
                code = caught_errors[0].get_error_code()
            elif caught_errors and isinstance(caught_errors[-1], TimeoutException):
                code = errors.STAGEINTIMEOUT if self.mode == 'stage-in' else errors.STAGEOUTTIMEOUT  # is it stage-in/out?
                self.logger.warning('caught time-out exception: %s' % caught_errors[0])
            else:
                code = ErrorCodes.STAGEINFAILED
            self.logger.fatal('caught_errors=%s' % str(caught_errors))
            self.logger.fatal('code=%s' % str(code))
            raise PilotException('failed to transfer files using copytools=%s, error=%s' % (copytools, caught_errors), code=code)

        self.logger.debug('result=%s' % str(result))
        return result


class StageInClient(StagingClient):

    mode = "stage-in"

    def resolve_replica(self, fspec, primary_schemas=None, allowed_schemas=None):
        """
            Resolve input replica first according to `primary_schemas`,
            if not found then look up within `allowed_schemas`
            :param fspec: input `FileSpec` objects
            :param allowed_schemas: list of allowed schemas or any if None
            :return: dict(surl, ddmendpoint, pfn)
            :raise PilotException: if replica not found
        """

        if not fspec.replicas:
            self.logger.warning('resolve_replicas() recevied no fspec.replicas')
            return

        allowed_schemas = allowed_schemas or [None]
        replica = None

        for ddmendpoint, replicas in fspec.replicas:
            if not replicas:  # ignore ddms with no replicas
                continue
            if primary_schemas:  ## look up primary schemas if requested
                replica = self.get_preferred_replica(replicas, primary_schemas)
            if not replica:
                replica = self.get_preferred_replica(replicas, allowed_schemas)
            if replica:
                surl = self.get_preferred_replica(replicas, ['srm']) or replicas[0]  # prefer SRM protocol for surl -- to be verified
                self.logger.info("[stage-in] surl (srm replica) from Rucio: pfn=%s, ddmendpoint=%s" % (surl, ddmendpoint))
                break

        if not replica:  # replica not found
            schemas = 'any' if not allowed_schemas[0] else ','.join(allowed_schemas)
            error = 'Failed to find replica for input file=%s, allowed_schemas=%s, fspec=%s' % (fspec.lfn, schemas, fspec)
            self.logger.error("resolve_replica: %s" % error)
            raise PilotException(error, code=ErrorCodes.REPLICANOTFOUND)

        return {'surl': surl, 'ddmendpoint': ddmendpoint, 'pfn': replica}

    def get_direct_access_variables(self, job):
        """
        Return the direct access settings for the PQ.

        :param job: job object.
        :return: allow_direct_access (bool), direct_access_type (string).
        """

        allow_direct_access, direct_access_type = False, ''
        if self.infosys.queuedata:  ## infosys is initialized
            allow_direct_access = self.infosys.queuedata.direct_access_lan or self.infosys.queuedata.direct_access_wan
            if self.infosys.queuedata.direct_access_lan:
                direct_access_type = 'LAN'
            if self.infosys.queuedata.direct_access_wan:
                direct_access_type = 'WAN'
        else:
            self.logger.info('infosys.queuedata is not initialized: direct access mode will be DISABLED by default')

        if job and not job.is_analysis() and job.transfertype != 'direct':  ## task forbids direct access
            allow_direct_access = False
            self.logger.info('switched off direct access mode for production job since transfertype=%s' % job.transfertype)

        return allow_direct_access, direct_access_type

    def set_accessmodes_for_direct_access(self, files, direct_access_type):
        """
        Update the FileSpec accessmodes for direct access and sort the files to get candidates for remote_io coming
        first in order to exclude them from checking of available space for stage-in.

        :param files: FileSpec objects.
        :param direct_access_type: type of direct access (LAN or WAN) (string).
        :return:
        """

        # sort the files
        files = sorted(files, key=lambda x: x.is_directaccess(ensure_replica=False), reverse=True)

        # populate allowremoteinputs for each FileSpec object
        for fdata in files:
            is_directaccess = fdata.is_directaccess(ensure_replica=False)
            if is_directaccess and direct_access_type == 'WAN':  ## is it the same for ES workflow ?? -- test and verify/FIXME LATER
                fdata.allowremoteinputs = True
            self.logger.info("check direct access for lfn=%s: allow_direct_access=true, fdata.is_directaccess()=%s =>"
                             " is_directaccess=%s, allowremoteinputs=%s" % (fdata.lfn,
                                                                            fdata.is_directaccess(ensure_replica=False),
                                                                            is_directaccess, fdata.allowremoteinputs))
            # must update accessmode for user jobs (it is only set already for production jobs)
            if fdata.accessmode != 'direct' and is_directaccess and fdata.accessmode != 'copy':
                fdata.accessmode = 'direct'

            # reset accessmode if direct access is not to be used
            if fdata.accessmode == 'direct' and not is_directaccess:
                fdata.accessmode = ''

            self.logger.info('accessmode for LFN=%s: %s (is_directaccess=%s)' %
                             (fdata.lfn, fdata.accessmode, is_directaccess))

    def transfer_files(self, copytool, files, activity=None, **kwargs):
        """
        Automatically stage in files using the selected copy tool module.

        :param copytool: copytool module
        :param files: list of `FileSpec` objects
        :param kwargs: extra kwargs to be passed to copytool transfer handler

        :return: the output of the copytool transfer operation
        :raise: PilotException in case of controlled error
        """

        #logger.debug('trace_report[eventVersion]=%s' % self.trace_report.get('eventVersion', 'unknown'))

        # sort out direct access logic
        job = kwargs.get('job', None)
        #job_access_mode = job.accessmode if job else ''
        allow_direct_access, direct_access_type = self.get_direct_access_variables(job)
        self.logger.info("direct access settings for the PQ: allow_direct_access=%s (type=%s)" %
                         (allow_direct_access, direct_access_type))
        kwargs['allow_direct_access'] = allow_direct_access

        if allow_direct_access:
            self.set_accessmodes_for_direct_access(files, direct_access_type)

        if getattr(copytool, 'require_replicas', False) and files and files[0].replicas is None:
            files = self.resolve_replicas(files)
            allowed_schemas = getattr(copytool, 'allowed_schemas', None)

            if self.infosys and self.infosys.queuedata:
                copytool_name = copytool.__name__.rsplit('.', 1)[-1]
                allowed_schemas = self.infosys.queuedata.resolve_allowed_schemas(activity, copytool_name) or allowed_schemas

            for fspec in files:
                resolve_replica = getattr(copytool, 'resolve_replica', None)
                if not callable(resolve_replica):
                    resolve_replica = self.resolve_replica

                ## prepare schemas which will be used to look up first the replicas allowed for direct access mode
                primary_schemas = self.direct_localinput_allowed_schemas if fspec.accessmode == 'direct' else None
                r = resolve_replica(fspec, primary_schemas, allowed_schemas)
                if not r:
                    raise ReplicasNotFound('resolve_replica() returned no replicas')

                if r.get('pfn'):
                    fspec.turl = r['pfn']
                if r.get('surl'):
                    fspec.surl = r['surl']  # TO BE CLARIFIED if it's still used and need
                if r.get('ddmendpoint'):
                    fspec.ddmendpoint = r['ddmendpoint']

                self.logger.info("[stage-in] found replica to be used for lfn=%s: ddmendpoint=%s, pfn=%s" %
                                 (fspec.lfn, fspec.ddmendpoint, fspec.turl))

        if not copytool.is_valid_for_copy_in(files):
            msg = 'input is not valid for transfers using copytool=%s' % copytool
            self.logger.warning(msg)
            self.logger.debug('input: %s' % files)
            self.trace_report.update(clientState='NO_REPLICA', stateReason=msg)
            self.trace_report.send()
            raise PilotException('invalid input data for transfer operation')

        if self.infosys:
            if self.infosys.queuedata:
                kwargs['copytools'] = self.infosys.queuedata.copytools
            kwargs['ddmconf'] = self.infosys.resolve_storage_data()
        kwargs['activity'] = activity

        # mark direct access files with status=remote_io
        if allow_direct_access:
            self.set_status_for_direct_access(files)

        # verify file sizes and available space for stage-in
        self.check_availablespace([e for e in files if e.status not in ['remote_io', 'transferred']])

        # add the trace report
        kwargs['trace_report'] = self.trace_report
        self.logger.info('ready to transfer (stage-in) files: %s' % files)

        return copytool.copy_in(files, **kwargs)

    def set_status_for_direct_access(self, files):
        """
        Update the FileSpec status with 'remote_io' for direct access mode.

        :param files: FileSpec objects.
        :return:
        """

        for fspec in files:
            if fspec.is_directaccess(ensure_replica=False):
                fspec.status_code = 0
                fspec.status = 'remote_io'

    def check_availablespace(self, files):
        """
        Verify that enough local space is available to stage in and run the job

        :param files: list of FileSpec objects.
        :raise: PilotException in case of not enough space or total input size too large
        """

        for f in files:
            self.logger.debug('lfn=%s filesize=%d accessmode=%s' % (f.lfn, f.filesize, f.accessmode))

        maxinputsize = convert_mb_to_b(get_maximum_input_sizes())
        totalsize = reduce(lambda x, y: x + y.filesize, files, 0)

        # verify total filesize
        if maxinputsize and totalsize > maxinputsize:
            error = "too many/too large input files (%s). total file size=%s B > maxinputsize=%s B" % \
                    (len(files), totalsize, maxinputsize)
            raise SizeTooLarge(error)

        self.logger.info("total input file size=%s B within allowed limit=%s B (zero value means unlimited)" %
                         (totalsize, maxinputsize))

        # get available space
        available_space = convert_mb_to_b(get_local_disk_space(os.getcwd()))
        self.logger.info("locally available space: %d B" % available_space)

        # are we within the limit?
        if totalsize > available_space:
            error = "not enough local space for staging input files and run the job (need %d B, but only have %d B)" % \
                    (totalsize, available_space)
            raise NoLocalSpace(error)


class StageOutClient(StagingClient):

    mode = "stage-out"

    def resolve_protocols(self, files, activity):
        """
            Populates filespec.protocols for each entry from `files` according to requested `activity`
            :param files: list of `FileSpec` objects
            :param activity: ordered list of preferred activity names to resolve SE protocols
            fdat.protocols = [dict(endpoint, path, flavour), ..]
            :return: `files`
        """

        ddmconf = self.infosys.resolve_storage_data()

        if isinstance(activity, basestring):
            activity = [activity]

        for fdat in files:
            ddm = ddmconf.get(fdat.ddmendpoint)
            if not ddm:
                raise Exception("Failed to resolve output ddmendpoint by name=%s (from PanDA), please check configuration. fdat=%s" % (fdat.ddmendpoint, fdat))

            protocols = []
            for aname in activity:
                aname = get_ddm_activity(aname)
                protocols = ddm.arprotocols.get(aname)
                if protocols:
                    break

            fdat.protocols = protocols

        return files

    def resolve_protocol(self, fspec, allowed_schemas=None):
        """
            Resolve protocols according to allowed schema
            :param fspec: `FileSpec` instance
            :param allowed_schemas: list of allowed schemas or any if None
            :return: list of dict(endpoint, path, flavour)
        """

        if not fspec.protocols:
            return []

        protocols = []

        allowed_schemas = allowed_schemas or [None]
        for schema in allowed_schemas:
            for pdat in fspec.protocols:
                if schema is None or pdat.get('endpoint', '').startswith("%s://" % schema):
                    protocols.append(pdat)

        return protocols

    def prepare_destinations(self, files, activities):
        """
            Resolve destination RSE (filespec.ddmendpoint) for each entry from `files` according to requested `activities`
            Apply Pilot-side logic to choose proper destination
            :param files: list of FileSpec objects to be processed
            :param activities: ordered list of activities to be used to resolve astorages
            :return: updated fspec entries
        """

        if not self.infosys.queuedata:  ## infosys is not initialized: not able to fix destination if need, nothing to do
            return files

        if isinstance(activities, (str, unicode)):
            activities = [activities]

        if not activities:
            raise PilotException("Failed to resolve destination: passed empty activity list. Internal error.",
                                 code=ErrorCodes.INTERNALPILOTPROBLEM, state='INTERNAL_ERROR')

        astorages = self.infosys.queuedata.astorages or {}

        storages = None
        activity = activities[0]
        for a in activities:
            storages = astorages.get(a, {})
            if storages:
                break

        if not storages:
            raise PilotException("Failed to resolve destination: no associated storages defined for activity=%s (%s)"
                                 % (activity, ','.join(activities)), code=ErrorCodes.NOSTORAGE, state='NO_ASTORAGES_DEFINED')

        # take the fist choice for now, extend the logic later if need
        ddm = storages[0]

        self.logger.info("[prepare_destinations][%s]: allowed (local) destinations: %s" % (activity, storages))
        self.logger.info("[prepare_destinations][%s]: resolved default destination ddm=%s" % (activity, ddm))

        for e in files:
            if not e.ddmendpoint:  ## no preferences => use default destination
                self.logger.info("[prepare_destinations][%s]: fspec.ddmendpoint is not set for lfn=%s"
                                 " .. will use default ddm=%s as (local) destination" % (activity, e.lfn, ddm))
                e.ddmendpoint = ddm
            elif e.ddmendpoint not in storages:  ## fspec.ddmendpoint is not in associated storages => assume it as final (non local) alternative destination
                self.logger.info("[prepare_destinations][%s]: Requested fspec.ddmendpoint=%s is not in the list of allowed (local) destinations"
                                 " .. will consider default ddm=%s for transfer and tag %s as alt. location" % (activity, e.ddmendpoint, ddm, e.ddmendpoint))
                e.ddmendpoint = ddm
                e.ddmendpoint_alt = e.ddmendpoint  ###  consider me later

        return files

    @classmethod
    def get_path(self, scope, lfn, prefix='rucio'):
        """
            Construct a partial Rucio PFN using the scope and the LFN
        """

        # <prefix=rucio>/<scope>/md5(<scope>:<lfn>)[0:2]/md5(<scope:lfn>)[2:4]/<lfn>

        hash_hex = hashlib.md5('%s:%s' % (scope, lfn)).hexdigest()

        #paths = [prefix] + scope.split('.') + [hash_hex[0:2], hash_hex[2:4], lfn]
        # exclude prefix from the path: this should be properly considered in protocol/AGIS for today
        paths = scope.split('.') + [hash_hex[0:2], hash_hex[2:4], lfn]
        paths = filter(None, paths)  # remove empty parts to avoid double /-chars

        return '/'.join(paths)

    def resolve_surl(self, fspec, protocol, ddmconf, **kwargs):
        """
            Get final destination SURL for file to be transferred
            Can be customized at the level of specific copytool
            :param protocol: suggested protocol
            :param ddmconf: full ddmconf data
            :param activity: ordered list of preferred activity names to resolve SE protocols
            :return: dict with keys ('pfn', 'ddmendpoint')
        """

        # consider only deterministic sites (output destination)

        ddm = ddmconf.get(fspec.ddmendpoint)
        if not ddm:
            raise PilotException('Failed to resolve ddmendpoint by name=%s' % fspec.ddmendpoint)

        # path = protocol.get('path', '').rstrip('/')
        # if not (ddm.is_deterministic or (path and path.endswith('/rucio'))):
        if not ddm.is_deterministic:
            raise PilotException('resolve_surl(): Failed to construct SURL for non deterministic ddm=%s: '
                                 'NOT IMPLEMENTED' % fspec.ddmendpoint, code=ErrorCodes.NONDETERMINISTICDDM)

        surl = protocol.get('endpoint', '') + os.path.join(protocol.get('path', ''), self.get_path(fspec.scope, fspec.lfn))
        return {'surl': surl}

    def transfer_files(self, copytool, files, activity, **kwargs):  # noqa: C901
        """
            Automatically stage out files using the selected copy tool module.

            :param copytool: copytool module
            :param files: list of `FileSpec` objects
            :param activity: ordered list of preferred activity names to resolve SE protocols
            :param kwargs: extra kwargs to be passed to copytool transfer handler

            :return: the output of the copytool transfer operation
            :raise: PilotException in case of controlled error
        """

        # check if files exist before actual processing
        # populate filesize if need, calc checksum
        for fspec in files:
            pfn = fspec.surl or getattr(fspec, 'pfn', None) or os.path.join(kwargs.get('workdir', ''), fspec.lfn)
            if not os.path.isfile(pfn) or not os.access(pfn, os.R_OK):
                msg = "Error: output pfn file does not exist: %s" % pfn
                self.logger.error(msg)
                self.trace_report.update(clientState='NO_REPLICA', stateReason=msg)
                self.trace_report.send()
                raise PilotException(msg, code=ErrorCodes.MISSINGOUTPUTFILE, state="FILE_INFO_FAIL")
            if not fspec.filesize:
                fspec.filesize = os.path.getsize(pfn)

            if not fspec.filesize:
                msg = 'output file has size zero: %s' % fspec.lfn
                self.logger.fatal(msg)
                raise PilotException(msg, code=ErrorCodes.ZEROFILESIZE, state="ZERO_FILE_SIZE")

            fspec.surl = pfn
            fspec.activity = activity
            if not fspec.checksum.get('adler32'):
                fspec.checksum['adler32'] = calculate_checksum(pfn)

        # prepare files (resolve protocol/transfer url)
        if getattr(copytool, 'require_protocols', True) and files:

            ddmconf = self.infosys.resolve_storage_data()
            allowed_schemas = getattr(copytool, 'allowed_schemas', None)

            if self.infosys and self.infosys.queuedata:
                copytool_name = copytool.__name__.rsplit('.', 1)[-1]
                allowed_schemas = self.infosys.queuedata.resolve_allowed_schemas(activity, copytool_name) or allowed_schemas

            files = self.resolve_protocols(files, activity)

            for fspec in files:

                protocols = self.resolve_protocol(fspec, allowed_schemas)
                if not protocols:  #  no protocols found
                    error = 'Failed to resolve protocol for file=%s, allowed_schemas=%s, fspec=%s' % (fspec.lfn, allowed_schemas, fspec)
                    self.logger.error("resolve_protocol: %s" % error)
                    raise PilotException(error, code=ErrorCodes.NOSTORAGEPROTOCOL)

                # take first available protocol for copytool: FIX ME LATER if need (do iterate over all allowed protocols?)
                protocol = protocols[0]

                self.logger.info("resolved protocol to be used for transfer: data=%s" % protocol)

                resolve_surl = getattr(copytool, 'resolve_surl', None)
                if not callable(resolve_surl):
                    resolve_surl = self.resolve_surl

                r = resolve_surl(fspec, protocol, ddmconf, activity=activity)  ## pass ddmconf & activity for possible custom look up at the level of copytool
                if r.get('surl'):
                    fspec.turl = r['surl']
                if r.get('ddmendpoint'):
                    fspec.ddmendpoint = r['ddmendpoint']

        if not copytool.is_valid_for_copy_out(files):
            self.logger.warning('Input is not valid for transfers using copytool=%s' % copytool)
            self.logger.debug('Input: %s' % files)
            raise PilotException('Invalid input for transfer operation')

        self.logger.info('ready to transfer (stage-out) files: %s' % files)

        if self.infosys:
            kwargs['copytools'] = self.infosys.queuedata.copytools

            # some copytools will need to know endpoint specifics (e.g. the space token) stored in ddmconf, add it
            kwargs['ddmconf'] = self.infosys.resolve_storage_data()

        if not files:
            msg = 'nothing to stage-out - an internal Pilot error has occurred'
            self.logger.fatal(msg)
            raise PilotException(msg, code=errors.INTERNALPILOTPROBLEM)

        # add the trace report
        kwargs['trace_report'] = self.trace_report

        return copytool.copy_out(files, **kwargs)

#class StageInClientAsync(object):
#
#    def __init__(self, site):
#        raise NotImplementedError
#
#    def queue(self, files):
#        raise NotImplementedError
#
#    def is_transferring(self):
#        raise NotImplementedError
#
#    def start(self):
#        raise NotImplementedError
#
#    def finish(self):
#        raise NotImplementedError
#
#    def status(self):
#        raise NotImplementedError
#
#
#class StageOutClientAsync(object):
#
#    def __init__(self, site):
#        raise NotImplementedError
#
#    def queue(self, files):
#        raise NotImplementedError
#
#    def is_transferring(self):
#        raise NotImplementedError
#
#    def start(self):
#        raise NotImplementedError
#
#    def finish(self):
#        raise NotImplementedError
#
#    def status(self):
#        raise NotImplementedError
